#include "m3hStd.h"
#include "m3iMesh.h"
#include "m3sMeshData.h"
#include "m3cPhysicsWorld.h"
#include "m3cPhysicsBody.h"
#include "m3iPhysicsWorldListener.h"

namespace melissa
{

m3cPhysicsWorld::m3cPhysicsWorld(m3iPhysicsWorldListener *listener)
    : m_listener(listener)
{
    m_broadphase = new btDbvtBroadphase();
    m_collisionConfiguration = new btDefaultCollisionConfiguration();
    m_dispatcher = new btCollisionDispatcher(m_collisionConfiguration);
    m_solver = new btSequentialImpulseConstraintSolver();
    m_world = new btDiscreteDynamicsWorld(m_dispatcher, m_broadphase, m_solver, m_collisionConfiguration);
    m_world->setGravity(btVector3(0, -9.8, 0));
}

m3cPhysicsWorld::~m3cPhysicsWorld()
{
    m_dontUnregisterBodies = true;
    for (auto b : m_registeredBodies)
        delete b;

    for (auto t : m_cachedTriMeshes)
        delete t;

    delete m_world;
    delete m_solver;
    delete m_dispatcher;
    delete m_collisionConfiguration;
    delete m_broadphase;
}

m3iPhysicsBody *m3cPhysicsWorld::createBoxBody(float mass, const glm::vec3 &dims)
{
    btCollisionShape* shape = new btBoxShape(btVector3(dims.x, dims.y, dims.z));
    return new m3cPhysicsBody(this, shape, mass);
}

m3iPhysicsBody *m3cPhysicsWorld::createSphereBody(float mass, float radius)
{
    btCollisionShape* shape = new btSphereShape(radius);
    return new m3cPhysicsBody(this, shape, mass);
}

m3iPhysicsBody *m3cPhysicsWorld::createMeshBody(float mass, const m3iMesh *mesh)
{
    btTriangleMesh *triMesh = new btTriangleMesh();
    m_cachedTriMeshes.push_back(triMesh);

    auto vertices = mesh->getData().vertices;
    auto elements = mesh->getData().elements;

    if (mesh->getData().elementsCount)
    {
        for (uint64_t i = 0; i < mesh->getData().elementsCount / 3; ++i)
        {
            btVector3 btV[3];

            for (uint8_t j = 0; j < 3; ++j)
            {
                const m3sVertex &v = vertices[elements[i * 3 + j]];
                btV[j] = btVector3(v.position.x, v.position.y, v.position.z);
            }

            triMesh->addTriangle(btV[0], btV[1], btV[2]);
        }
    }
    else
    {
        for (uint64_t i = 0; i < mesh->getData().verticesCount / 3; ++i)
        {
            btVector3 btV[3];

            for (uint8_t j = 0; j < 3; ++j)
            {
                const m3sVertex &v = vertices[i * 3 + j];
                btV[j] = btVector3(v.position.x, v.position.y, v.position.z);
            }

            triMesh->addTriangle(btV[0], btV[1], btV[2]);
        }
    }

    btBvhTriangleMeshShape *shape = new btBvhTriangleMeshShape(triMesh, true);
    return new m3cPhysicsBody(this, shape, mass);
}

void m3cPhysicsWorld::simulate(float dt)
{
    m_world->stepSimulation(dt, 2);
}

void m3cPhysicsWorld::registerPhysicsBody(m3iPhysicsBody *body)
{
    m_listener->addedPhysicsBody(body);
    m_registeredBodies.push_back(body);
    m_world->addRigidBody(static_cast<m3cPhysicsBody*>(body)->getRigidBody());
}

void m3cPhysicsWorld::unregisterPhysicsBody(m3iPhysicsBody *body)
{
    m_listener->removedPhysicsBody(body);
    m_world->removeRigidBody(static_cast<m3cPhysicsBody*>(body)->getRigidBody());
    if (m_dontUnregisterBodies)
        return;
    remove_from_vector(m_registeredBodies, body);
}

}
