#include "m3cLogger.h"
#include "m3cHardwareMesh.h"

namespace melissa
{

m3cHardwareMesh::m3cHardwareMesh(m3iResourceManager *resMan, const m3sMeshData &data, bool optimize)
    : m3cAbstractMesh(resMan, data),
      m_vertices(m_data.vertices, m_data.vertices + m_data.verticesCount),
      m_elements(m_data.elements, m_data.elements + m_data.elementsCount)
{
    if (optimize)
        this->optimize();
    reloadHandles();
}

m3cHardwareMesh::~m3cHardwareMesh()
{
    freeHandles();
}

void m3cHardwareMesh::bind()
{
    glBindBuffer(GL_ARRAY_BUFFER, m_vbo[0]);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_vbo[1]);
}

const void *m3cHardwareMesh::getVertexPointer() const
{
    return 0;
}

const void *m3cHardwareMesh::getElementPointer() const
{
    return 0;
}

void m3cHardwareMesh::freeHandles()
{
    glDeleteBuffers(2, m_vbo);
}

void m3cHardwareMesh::reloadHandles()
{
    glGenBuffers(2, m_vbo);
    if (m_vbo[0] == 0 || m_vbo[1] == 0)
    {
        m3cLogger::fatal() << "Can't create VBO";
        return;
    }
    bind();
    glBufferData(GL_ARRAY_BUFFER, m_vertices.size() * sizeof(m3sVertex), m_vertices.data(), GL_STATIC_DRAW);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, m_elements.size() * sizeof(m3sElement), m_elements.data(), GL_STATIC_DRAW);

    m_data.vertices = m_vertices.data();
    m_data.verticesCount = m_vertices.size();
    m_data.elements = m_elements.data();
    m_data.elementsCount = m_elements.size();
}

void m3cHardwareMesh::optimize()
{
    assert(m_data.useElements);

    std::set<std::string> names;
    for (const m3sMeshSurface &s : m_data.surfaces)
        names.insert(s.material.name);

    std::vector<m3sElement> newElements;
    std::vector<m3sMeshSurface> newSurfaces;

    for (auto it = names.begin(); it != names.end(); ++it)
    {
        m3sMeshSurface newSurface;
        newSurface.offset = newElements.size();
        newSurface.material.name = *it;

        for (const m3sMeshSurface &s : m_data.surfaces)
        {
            if (s.material.name != *it)
                continue;

            newSurface.material = s.material;

            for (uint64_t i = s.offset; i < s.offset + s.count; ++i)
            {
                m3sElement e = m_elements[i];
                newElements.push_back(e);
            }
        }

        newSurface.count = newElements.size() - newSurface.offset;
        newSurfaces.push_back(newSurface);
    }

    m_data.surfaces = newSurfaces;
    m_elements = newElements;
}

}
