#include "m3cLogger.h"
#include "m3cProgram.h"
#include "m3iResourceManager.h"

namespace melissa
{

m3cProgram *m3cProgram::s_program = 0;

m3cProgram::m3cProgram(m3iResourceManager *resMan, const std::string &source)
    : m_source(source),
      m_resMan(resMan)
{
    resMan->registerResource(this);
    reloadHandles();
}

m3cProgram::~m3cProgram()
{
    freeHandles();
    m_resMan->unregisterResource(this);
}

const m3sProgramHandles &m3cProgram::getProgramHandles() const
{
    return m_handles;
}

GLuint m3cProgram::createShaderFromSource(const std::string &source, GLenum type) const
{
#ifndef MELISSA_MOBILE
    std::string version = "#version 110\n";
#else
    std::string version = "precision mediump float;\n";
#endif
    std::string limits =
            "#define _MAX_JOINTS_ 48\n"
            "#define _MAX_LIGHTS_ 4\n"
    ;
    std::string lightStruct =
            "struct Light {\n"
            "   vec4 DIFFUSE;\n"
            "   vec3 POSITION;\n"
            "   vec3 DIRECTION;\n"
            "   vec3 ATTENUATION;\n"
            "};\n"
    ;
    std::string completeVsrc = version + limits + lightStruct + source;
    const char *vscstr = completeVsrc.c_str();

    GLuint shader = glCreateShader(type);
    glShaderSource(shader, 1, &vscstr, NULL);
    glCompileShader(shader);

    GLint isCompiled = 0;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &isCompiled);
    if(isCompiled == GL_FALSE)
    {
        GLint maxLength = 0;
        glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &maxLength);
        std::vector<char> errorLog(maxLength);
        glGetShaderInfoLog(shader, maxLength, &maxLength, &errorLog[0]);
        std::string shaderType = "vertex";
        if (type == GL_FRAGMENT_SHADER)
            shaderType = "fragment";
        m3cLogger::fatal() << shaderType << "shader compilation error \"" << errorLog.data() << "\"\n" << source;
    }

    return shader;
}

void m3cProgram::initializeHandles()
{
    m_handles.AttributePosition = glGetAttribLocation(m_program, "POSITION");
    m_handles.AttributeNormal = glGetAttribLocation(m_program, "NORMAL");
    m_handles.AttributeTangent = glGetAttribLocation(m_program, "TANGENT");
    m_handles.AttributeBitangent = glGetAttribLocation(m_program, "BITANGENT");
    m_handles.AttributePositionNext = glGetAttribLocation(m_program, "POSITION_NEXT");
    m_handles.AttributeNormalNext = glGetAttribLocation(m_program, "NORMAL_NEXT");
    m_handles.AttributeTexcoord1 = glGetAttribLocation(m_program, "TEXCOORD1");
    m_handles.AttributeTexcoord2 = glGetAttribLocation(m_program, "TEXCOORD2");
    m_handles.AttributeColor = glGetAttribLocation(m_program, "COLOR");
    m_handles.AttributeJoint = glGetAttribLocation(m_program, "JOINT");
    m_handles.AttributeWeight = glGetAttribLocation(m_program, "WEIGHT");

    m_handles.UniformM = glGetUniformLocation(m_program, "M");
    m_handles.UniformMv = glGetUniformLocation(m_program, "MV");
    m_handles.UniformMvp = glGetUniformLocation(m_program, "MVP");
    m_handles.UniformShadowMapBiasMatrix = glGetUniformLocation(m_program, "SHADOWMAP_BIAS_MATRIX");
    m_handles.UniformNormalMatrix = glGetUniformLocation(m_program, "NORMAL_MATRIX");
    m_handles.UniformTexture = glGetUniformLocation(m_program, "TEXTURE");
    m_handles.UniformNormalMapTexture = glGetUniformLocation(m_program, "NORMALMAP_TEXTURE");
    m_handles.UniformLightMapTexture = glGetUniformLocation(m_program, "LIGHTMAP_TEXTURE");
    m_handles.UniformShadowMapTexture = glGetUniformLocation(m_program, "SHADOWMAP_TEXTURE");
    m_handles.UniformScreenTexture = glGetUniformLocation(m_program, "SCREEN_TEXTURE");
    m_handles.UniformMaterialDiffuse = glGetUniformLocation(m_program, "MATERIAL_DIFFUSE");
    m_handles.UniformMaterialSpecular = glGetUniformLocation(m_program, "MATERIAL_SPECULAR");
    m_handles.UniformScreenSize = glGetUniformLocation(m_program, "SCREEN_SIZE");
    m_handles.UniformFrameTime = glGetUniformLocation(m_program, "FRAME_TIME");
    m_handles.UniformJointMatrix = glGetUniformLocation(m_program, "JOINT_MATRIX");
    m_handles.UniformFogColor = glGetUniformLocation(m_program, "FOG_COLOR");
    m_handles.UniformFogStart = glGetUniformLocation(m_program, "FOG_START");
    m_handles.UniformFogEnd = glGetUniformLocation(m_program, "FOG_END");

    m_handles.UniformCameraPosition = glGetUniformLocation(m_program, "CAMERA_POSITION");

    m_handles.UniformSunColor = glGetUniformLocation(m_program, "SUN_COLOR");
    m_handles.UniformSunPosition = glGetUniformLocation(m_program, "SUN_POSITION");
    m_handles.UniformSunDirection = glGetUniformLocation(m_program, "SUN_DIRECTION");

    m_handles.UniformLightsCount = glGetUniformLocation(m_program, "LIGHTS_COUNT");

    for (int i = 0; i < 4; ++i)
    {
        std::stringstream diffuseName;
        diffuseName << "LIGHT[" << i << "].DIFFUSE";

        std::stringstream positionName;
        positionName << "LIGHT[" << i << "].POSITION";

        std::stringstream directionName;
        directionName << "LIGHT[" << i << "].DIRECTION";

        std::stringstream attenuationName;
        attenuationName << "LIGHT[" << i << "].ATTENUATION";

        m_handles.UniformLightDiffuse[i] = glGetUniformLocation(m_program, diffuseName.str().c_str());
        m_handles.UniformLightPosition[i] = glGetUniformLocation(m_program, positionName.str().c_str());
        m_handles.UniformLightDirection[i] = glGetUniformLocation(m_program, directionName.str().c_str());
        m_handles.UniformLightAttenuation[i] = glGetUniformLocation(m_program, attenuationName.str().c_str());
    }
}

void m3cProgram::enableAttributes()
{
    glEnableVertexAttribArray(m_handles.AttributePosition);
    glEnableVertexAttribArray(m_handles.AttributeNormal);
    glEnableVertexAttribArray(m_handles.AttributeTangent);
    glEnableVertexAttribArray(m_handles.AttributeBitangent);
    glEnableVertexAttribArray(m_handles.AttributePositionNext);
    glEnableVertexAttribArray(m_handles.AttributeNormalNext);
    glEnableVertexAttribArray(m_handles.AttributeTexcoord1);
    glEnableVertexAttribArray(m_handles.AttributeTexcoord2);
    glEnableVertexAttribArray(m_handles.AttributeColor);
    glEnableVertexAttribArray(m_handles.AttributeJoint);
    glEnableVertexAttribArray(m_handles.AttributeWeight);
}

void m3cProgram::disableAttributes()
{
    glDisableVertexAttribArray(m_handles.AttributePosition);
    glDisableVertexAttribArray(m_handles.AttributeNormal);
    glDisableVertexAttribArray(m_handles.AttributeTangent);
    glDisableVertexAttribArray(m_handles.AttributeBitangent);
    glDisableVertexAttribArray(m_handles.AttributePositionNext);
    glDisableVertexAttribArray(m_handles.AttributeNormalNext);
    glDisableVertexAttribArray(m_handles.AttributeTexcoord1);
    glDisableVertexAttribArray(m_handles.AttributeTexcoord2);
    glDisableVertexAttribArray(m_handles.AttributeColor);
    glDisableVertexAttribArray(m_handles.AttributeJoint);
    glDisableVertexAttribArray(m_handles.AttributeWeight);
}

void m3cProgram::use()
{
    if (s_program != this)
    {
        if (s_program)
            s_program->disableAttributes();
        glUseProgram(m_program);
        enableAttributes();
    }
    s_program = this;
}

void m3cProgram::freeHandles()
{
    if (s_program && s_program->m_program == m_program)
        resetCurrent();
    glDeleteProgram(m_program);
}

void m3cProgram::reloadHandles()
{
    m_program = glCreateProgram();

    auto sv = createShaderFromSource("#define _VERTEX_\n" + m_source, GL_VERTEX_SHADER);
    auto sf = createShaderFromSource("#define _FRAGMENT_\n" + m_source, GL_FRAGMENT_SHADER);

    glAttachShader(m_program, sv);
    glAttachShader(m_program, sf);
    glLinkProgram(m_program);

    glDeleteShader(sv);
    glDeleteShader(sf);

    initializeHandles();
}

const uint8_t m3cProgram::getMaxLights()
{
    return 4;
}

const uint8_t m3cProgram::getMaxJoints()
{
    return 48;
}

void m3cProgram::resetCurrent()
{
    s_program = 0;
}

}
