#include "m3cLogger.h"
#include "m3cRender.h"
#include "m3cProgram.h"
#include "m3cTexture.h"
#include "m3iResource.h"
#include "m3iSkeleton.h"
#include "m3cLinkedMesh.h"
#include "m3cHardwareMesh.h"
#include "m3cTextureRenderTarget.h"

namespace
{

melissa::m3sVertex createVertex(float x, float y, float z, float u, float v)
{
    melissa::m3sVertex vert;
    vert.position.x = x;
    vert.position.y = y;
    vert.position.z = z;
    vert.texcoord1.x = u;
    vert.texcoord1.y = v;
    return vert;
}

}

namespace melissa
{

m3cRender::m3cRender()
{
    for (int i = 0; i < 2; ++i)
    for (int j = 0; j < 2; ++j)
    for (int k = 0; k < 2; ++k)
    for (int l = 0; l < 2; ++l)
    for (int m = 0; m < 2; ++m)
    for (int n = 0; n < 2; ++n)
    for (int o = 0; o < 2; ++o)
    for (int p = 0; p < 2; ++p)
    {
        m_builtInPrograms[i][j][k][l][m][n][o][p] = 0;
    }

    GLint mvv = 0;
    glGetIntegerv(GL_MAX_VERTEX_UNIFORM_VECTORS, &mvv);
    m3cLogger::info() << "GL_MAX_VERTEX_UNIFORM_VECTORS" << mvv; 

    m_builtIn2dProgram = createProgram(
                "uniform sampler2D TEXTURE;                                                     \n"
                "varying vec2 texcoord;                                                         \n"
                "                                                                               \n"
                "#ifdef _VERTEX_                                                                \n"
                "                                                                               \n"
                "attribute vec3 POSITION;                                                       \n"
                "attribute vec2 TEXCOORD1;                                                       \n"
                "                                                                               \n"
                "void main()                                                                    \n"
                "{                                                                              \n"
                "    gl_Position = vec4(POSITION, 1.0);                                         \n"
                "    texcoord = TEXCOORD1;                                                       \n"
                "}                                                                              \n"
                "                                                                               \n"
                "#endif                                                                         \n"
                "                                                                               \n"
                "#ifdef _FRAGMENT_                                                              \n"
                "                                                                               \n"
                "void main()                                                                    \n"
                "{                                                                              \n"
                "    gl_FragColor = texture2D(TEXTURE, texcoord);                               \n"
                "}                                                                              \n"
                "                                                                               \n"
                "#endif                                                                         \n"
    );

    // TOTO: morph animation
    m_depthProgram[false] = createProgram(
                "uniform mat4 MVP;                                                              \n"
                "                                                                               \n"
                "#ifdef _VERTEX_                                                                \n"
                "                                                                               \n"
                "attribute vec3 POSITION;                                                       \n"
                "                                                                               \n"
                "void main()                                                                    \n"
                "{                                                                              \n"
                "    gl_Position = MVP * vec4(POSITION, 1.0);                                   \n"
                "}                                                                              \n"
                "                                                                               \n"
                "#endif                                                                         \n"
                "                                                                               \n"
                "#ifdef _FRAGMENT_                                                              \n"
                "                                                                               \n"
                "void main()                                                                    \n"
                "{                                                                              \n"
                "    float fd = gl_FragCoord.z;                                                 \n"
                "    gl_FragColor = vec4(fd, fd, fd, 1.0);                                      \n"
                "}                                                                              \n"
                "                                                                               \n"
                "#endif                                                                         \n"
    );

    m_depthProgram[true] = createProgram(
                "uniform mat4 JOINT_MATRIX[_MAX_JOINTS_];                                       \n"
                "uniform mat4 MVP;                                                              \n"
                "                                                                               \n"
                "#ifdef _VERTEX_                                                                \n"
                "                                                                               \n"
                "attribute vec3 POSITION;                                                       \n"
                "attribute vec4 JOINT;                                                          \n"
                "attribute vec4 WEIGHT;                                                         \n"
                "                                                                               \n"
                "void main()                                                                    \n"
                "{                                                                              \n"
                "    mat4 newMat = JOINT_MATRIX[int(JOINT.x)] * WEIGHT.x;                       \n"
                "    newMat += JOINT_MATRIX[int(JOINT.y)] * WEIGHT.y;                           \n"
                "    newMat += JOINT_MATRIX[int(JOINT.z)] * WEIGHT.z;                           \n"
                "    newMat += JOINT_MATRIX[int(JOINT.w)] * WEIGHT.w;                           \n"
                "    gl_Position = MVP * (newMat * vec4(POSITION, 1.0));                        \n"
                "}                                                                              \n"
                "                                                                               \n"
                "#endif                                                                         \n"
                "                                                                               \n"
                "#ifdef _FRAGMENT_                                                              \n"
                "                                                                               \n"
                "void main()                                                                    \n"
                "{                                                                              \n"
                "    float fd = gl_FragCoord.z;                                                 \n"
                "    gl_FragColor = vec4(fd, fd, fd, 1.0);                                      \n"
                "}                                                                              \n"
                "                                                                               \n"
                "#endif                                                                         \n"
    );

    m_shadowMapTexture = reinterpret_cast<m3cTexture*>(createTexture(1024, 1024, 0, 0));
    m_shadowsRenderTarget = createTextureRenderTarget(m_shadowMapTexture);

    const m3sVertex vertices[4] = {
        createVertex(-1, -1, 0, 0, 0),
        createVertex(-1, 1, 0, 0, 1),
        createVertex(1, 1, 0, 1, 1),
        createVertex(1, -1, 0, 1, 0)
    };
    const m3sElement elements[6] = {0, 1, 2, 0, 2, 3};

    m3sMeshSurface surface;
    surface.count = 6;

    m3sMeshData meshData;
    meshData.vertices = vertices;
    meshData.verticesCount = 4;
    meshData.elements = elements;
    meshData.elementsCount = 6;
    meshData.surfaces.push_back(surface);
    m_postPlaneMesh = static_cast<m3cAbstractMesh*>(createHardwareMesh(meshData, false));

    m_clearEffect.prog = createProgram(
                "uniform sampler2D TEXTURE;                                                     \n"
                "varying vec2 texcoord;                                                         \n"
                "                                                                               \n"
                "#ifdef _VERTEX_                                                                \n"
                "                                                                               \n"
                "attribute vec3 POSITION;                                                       \n"
                "attribute vec2 TEXCOORD1;                                                       \n"
                "                                                                               \n"
                "void main()                                                                    \n"
                "{                                                                              \n"
                "    gl_Position = vec4(POSITION, 1.0);                                         \n"
                "    texcoord = TEXCOORD1;                                                       \n"
                "}                                                                              \n"
                "                                                                               \n"
                "#endif                                                                         \n"
                "                                                                               \n"
                "#ifdef _FRAGMENT_                                                              \n"
                "                                                                               \n"
                "void main()                                                                    \n"
                "{                                                                              \n"
                "    gl_FragColor = texture2D(TEXTURE, texcoord);                               \n"
                "}                                                                              \n"
                "                                                                               \n"
                "#endif                                                                         \n"
    );
    m_clearEffect.tex = reinterpret_cast<m3cTexture*>(createTexture(0, 0, 0, 0));
    m_clearEffect.rt = createTextureRenderTarget(m_clearEffect.tex);

    m_currentPostEffectSet = createPostEffectSet();
}

m3cRender::~m3cRender()
{
    m_dontUnregisterHandles = true;
    delete_all(m_registeredResources);
}

m3iMesh *m3cRender::createHardwareMesh(const m3sMeshData &data, bool optimize)
{
    return new m3cHardwareMesh(this, data, optimize);
}

m3iMesh *m3cRender::createLinkedMesh(const m3sMeshData &data)
{
    return new m3cLinkedMesh(this, data);
}

void m3cRender::setFrameData(const m3sRenderFrameData &data)
{
    m_frameData = data;
}

void m3cRender::clear()
{
    m_meshesForDraw.clear();
}

void m3cRender::drawMesh(const m3sDrawMeshData &data)
{
    m_meshesForDraw.push_back(data);
}

void m3cRender::flush()
{  
    m_frustum.build(m_frameData.projectionMatrix, m_frameData.viewMatrix);
    burnShadowMap();
    renderScene();
}

uint64_t m3cRender::createPostEffectSet()
{
    m_postEffectSets.push_back(PostEffectSet());
    return m_postEffectSets.size() - 1;
}

void m3cRender::clearPostEffectSet(uint64_t id)
{
    m_postEffectSets[id].postEffects.clear();
}

uint64_t m3cRender::getCurrentPostEffectSet() const
{
    return m_currentPostEffectSet;
}

void m3cRender::usePostEffectSet(uint64_t id)
{
    m_currentPostEffectSet = id;
}

void m3cRender::renderScene()
{
    auto lastRt = m_clearEffect.rt;
    m3sDrawMeshData postPlaneDrawData;
    postPlaneDrawData.frustumEnabled = false;
    postPlaneDrawData.depth = false;
    postPlaneDrawData.mesh = m_postPlaneMesh;
    auto vp = getViewport();

    std::vector<PostEffect> &postEffects = m_postEffectSets[m_currentPostEffectSet].postEffects;

    if (postEffects.size())
        m_clearEffect.rt->bind();

    glClearColor(m_frameData.clearColor.r, m_frameData.clearColor.g,
                 m_frameData.clearColor.b, m_frameData.clearColor.a);
    glDepthMask(GL_TRUE);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    m_pvMatrix = m_frameData.projectionMatrix * m_frameData.viewMatrix;
    for (auto m : m_meshesForDraw)
        renderMesh(m);

    for (uint64_t i = 0; i < postEffects.size(); ++i)
    {
        postEffects[i].prog->use();
        m_clearEffect.tex->bind(8);
        glUniform1i(postEffects[i].prog->getProgramHandles().UniformScreenTexture, 8);

        glViewport(0, 0, postEffects[i].tex->getWidth(), postEffects[i].tex->getHeight());
        postEffects[i].rt->bind();
        m_postPlaneMesh->editSurfaces()[0].material.diffuseTexture = lastRt->getTexture();
        renderMesh(postPlaneDrawData, postEffects[i].prog);
        lastRt = postEffects[i].rt;
    }

    if (postEffects.size())
    {
        glViewport(vp.x, vp.y, vp.z, vp.w);
        m3cTextureRenderTarget::unbind();
        m_postPlaneMesh->editSurfaces()[0].material.diffuseTexture = lastRt->getTexture();
        renderMesh(postPlaneDrawData, m_clearEffect.prog);
    }
}

void m3cRender::burnShadowMap()
{
    if (!m_frameData.shadowsData.enabled)
        return;

    auto sunp = m_frameData.sun.position;
    auto sund = m_frameData.sun.direction;
    glm::mat4 sunViewMatrix = glm::lookAt(sunp, sunp + sund, glm::vec3(0, 1, 0));

    glClearColor(1, 1, 1, 1);
    float s = m_frameData.shadowsData.scale;
    m_pvMatrix = glm::ortho(-s, s, -s, s, -s, s) * sunViewMatrix;

    m_pvShadowsBiasMatrix = m_biasMatrix * m_pvMatrix;

    glm::vec4 lastVp = getViewport();
    glViewport(0, 0, m_shadowMapTexture->getWidth(), m_shadowMapTexture->getHeight());

    m_shadowsRenderTarget->bind();
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    for (auto m : m_meshesForDraw)
    {
        if (m.castShadows)
            renderMesh(m, m_depthProgram[static_cast<m3cAbstractMesh*>(m.mesh)->getData().hasSkeletalAnim]);
    }

    m3cTextureRenderTarget::unbind();
    glViewport(0, 0, lastVp.z, lastVp.w);
}

glm::vec4 m3cRender::getViewport() const
{
    glm::vec4 result;
    GLint vp[4] = {0};
    glGetIntegerv(GL_VIEWPORT, vp);
    result.x = vp[0];
    result.y = vp[1];
    result.z = vp[2];
    result.w = vp[3];
    return result;
}

m3cProgram *m3cRender::getBuiltInProgram(bool fog, bool tex, bool norm, bool lm, bool lig, bool shad, bool morph, bool skel)
{
    if (m_builtInPrograms[fog][tex][norm][lm][lig][shad][morph][skel])
        return m_builtInPrograms[fog][tex][norm][lm][lig][shad][morph][skel];

    const std::string source =
            "uniform mat4 M;                                                                \n"
            "uniform mat4 MV;                                                               \n"
            "uniform mat4 MVP;                                                              \n"
            "varying vec3 mPosition;                                                        \n"
            "varying vec3 mvPosition;                                                       \n"
            "varying vec4 color;                                                            \n"
            "                                                                               \n"
            "#ifdef _TEXTURED_                                                              \n"
            "    uniform sampler2D TEXTURE;                                                 \n"
            "    varying vec2 texcoord;                                                     \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _LIGHTMAP_                                                              \n"
            "    uniform sampler2D LIGHTMAP_TEXTURE;                                        \n"
            "    varying vec2 texcoord2;                                                    \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _SKELETAL_                                                              \n"
            "    uniform mat4 JOINT_MATRIX[_MAX_JOINTS_];                                   \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _LIGHTING_                                                              \n"
            "    uniform Light LIGHT[_MAX_LIGHTS_];                                         \n"
            "    uniform mat3 NORMAL_MATRIX;                                                \n"
            "    uniform vec3 SUN_DIRECTION;                                                \n"
            "    uniform vec4 SUN_COLOR;                                                    \n"
            "    uniform float LIGHTS_COUNT;                                                \n"
            "    varying vec3 normal;                                                       \n"
            "    varying float sunFactor;                                                   \n"
            "#ifdef _NORMALMAPPING_                                                         \n"
            "    uniform sampler2D NORMALMAP_TEXTURE;                                       \n"
            "    varying vec3 tangent;                                                      \n"
            "    varying vec3 bitangent;                                                    \n"
            "#endif                                                                         \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _SHADOWS_                                                               \n"
            "    uniform sampler2D SHADOWMAP_TEXTURE;                                       \n"
            "    uniform mat4 SHADOWMAP_BIAS_MATRIX;                                        \n"
            "    varying vec4 shadowmap_coord;                                              \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _FOG_                                                                   \n"
            "    uniform vec4 FOG_COLOR;                                                    \n"
            "    uniform float FOG_START;                                                   \n"
            "    uniform float FOG_END;                                                     \n"
            "    varying float fogFactor;                                                   \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "// Vertex                                                                      \n"
            "#ifdef _VERTEX_                                                                \n"
            "                                                                               \n"
            "attribute vec3 POSITION;                                                       \n"
            "attribute vec4 COLOR;                                                          \n"
            "                                                                               \n"
            "#ifdef _LIGHTING_                                                              \n"
            "    attribute vec3 NORMAL;                                                     \n"
            "    attribute vec3 TANGENT;                                                    \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _TEXTURED_                                                              \n"
            "    attribute vec2 TEXCOORD1;                                                  \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _LIGHTMAP_                                                              \n"
            "    attribute vec2 TEXCOORD2;                                                  \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _SKELETAL_                                                              \n"
            "    attribute vec4 JOINT;                                                      \n"
            "    attribute vec4 WEIGHT;                                                     \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _LIGHTING_                                                              \n"
            "float calcLightFactor(vec3 lDir, vec3 n)                                       \n"
            "{                                                                              \n"
            "    float factor = max(dot(n, lDir), 0.0);                                     \n"
            "    return clamp(factor, 0.0, 1.0);                                            \n"
            "}                                                                              \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "void main()                                                                    \n"
            "{                                                                              \n"
            "    vec4 pos = vec4(POSITION, 1.0);                                            \n"
            "                                                                               \n"
            "#ifdef _LIGHTING_                                                              \n"
            "    normal = normalize(NORMAL_MATRIX * NORMAL);                                \n"
            "    sunFactor = calcLightFactor(-SUN_DIRECTION, normal);                       \n"
            "#ifdef _NORMALMAPPING_                                                         \n"
            "    tangent = normalize(NORMAL_MATRIX * TANGENT);                              \n"
            "    bitangent = cross(normal, tangent);                                        \n"
            "#endif                                                                         \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _SKELETAL_                                                              \n"
            "    mat4 newMat = JOINT_MATRIX[int(JOINT.x)] * WEIGHT.x;                       \n"
            "    newMat += JOINT_MATRIX[int(JOINT.y)] * WEIGHT.y;                           \n"
            "    newMat += JOINT_MATRIX[int(JOINT.z)] * WEIGHT.z;                           \n"
            "    newMat += JOINT_MATRIX[int(JOINT.w)] * WEIGHT.w;                           \n"
            "    pos = (newMat * pos);                                                      \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "    gl_Position = MVP * pos;                                                   \n"
            "                                                                               \n"
            "#ifdef _TEXTURED_                                                              \n"
            "    texcoord = TEXCOORD1;                                                      \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _LIGHTMAP_                                                              \n"
            "    texcoord2 = TEXCOORD2;                                                     \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _SHADOWS_                                                               \n"
            "    shadowmap_coord = SHADOWMAP_BIAS_MATRIX * pos;                             \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _FOG_                                                                   \n"
            "    fogFactor = clamp((FOG_END - -(MV * pos).z) / (FOG_END - FOG_START), 0.0, 1.0);\n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "    mPosition = (M * pos).xyz;                                                 \n"
            "    mvPosition = (MV * pos).xyz;                                               \n"
            "    color = COLOR;                                                             \n"
            "}                                                                              \n"
            "                                                                               \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "// Fragment                                                                    \n"
            "#ifdef _FRAGMENT_                                                              \n"
            "                                                                               \n"
            "#ifdef _LIGHTING_                                                              \n"
            "vec4 calcPointLightColor(Light light)                                          \n"
            "{                                                                              \n"
            "    vec3 lDir = (light.POSITION - mPosition);                                  \n"
            "    float dist = length(lDir);                                                 \n"
            "    float att = 1.0 / (light.ATTENUATION[0] + light.ATTENUATION[1] *           \n"
            "        dist + light.ATTENUATION[2] * dist * dist);                            \n"
            "#ifdef _NORMALMAPPING_                                                         \n"
            "    vec3 bump = normalize(texture2D(NORMALMAP_TEXTURE, texcoord).xyz           \n"
            "        * 2.0 - 1.0);                                                          \n"
            "    vec3 lightVec;                                                             \n"
            "    vec3 eyeVec;                                                               \n"
            "    lightVec.x = dot(lDir, tangent);                                           \n"
            "    lightVec.y = dot(lDir, bitangent);                                         \n"
            "    lightVec.z = dot(lDir, normal);                                            \n"
            "    vec3 tmpVec = -lDir;                                                       \n"
            "    eyeVec.x = dot(tmpVec, tangent);                                           \n"
            "    eyeVec.y = dot(tmpVec, bitangent);                                         \n"
            "    eyeVec.z = dot(tmpVec, normal);                                            \n"
            "    float distSqrt = dot(lightVec, lightVec);                                  \n"
            "    vec3 lVec = lightVec * inversesqrt(distSqrt);                              \n"
            "    float factor = max(dot(lVec, bump), 0.0);                                  \n"
            "    return light.DIFFUSE * factor * att;                                       \n"
            "#else                                                                          \n"
            "    float ndotl = max(dot(normal, normalize(lDir)), 0.0);                      \n"
            "    return light.DIFFUSE * ndotl * att;                                        \n"
            "#endif                                                                         \n"
            "}                                                                              \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "void main()                                                                    \n"
            "{                                                                              \n"
            "    vec4 finalColor = color;                                                   \n"
            "                                                                               \n"
            "#ifdef _TEXTURED_                                                              \n"
            "    finalColor *= texture2D(TEXTURE, texcoord);                                \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _LIGHTING_                                                              \n"
            "    vec4 pointLightsColor = vec4(0.0);                                         \n"
            "    for (int i = 0; i < int(LIGHTS_COUNT); ++i)                                \n"
            "    {                                                                          \n"
            "        pointLightsColor += calcPointLightColor(LIGHT[i]);                     \n"
            "    }                                                                          \n"
            "    finalColor *= pointLightsColor + sunFactor * SUN_COLOR;                    \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _SHADOWS_                                                               \n"
            "   if (texture2D(SHADOWMAP_TEXTURE, shadowmap_coord.xy).z                      \n"
            "       < shadowmap_coord.z - 0.005)                                            \n"
            "          finalColor *= 0.5;                                                   \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _FOG_                                                                   \n"
            "    finalColor = mix(FOG_COLOR, finalColor, fogFactor);                        \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "#ifdef _LIGHTMAP_                                                              \n"
            "    finalColor *= texture2D(LIGHTMAP_TEXTURE, texcoord2);                      \n"
            "#endif                                                                         \n"
            "                                                                               \n"
            "    gl_FragColor = finalColor;                                                 \n"
            "}                                                                              \n"
            "                                                                               \n"
            "#endif                                                                         \n"
    ;

    std::string definitions;
    if (fog) definitions += "#define _FOG_\n";
    if (tex) definitions += "#define _TEXTURED_\n";
    if (norm) definitions += "#define _NORMALMAPPING_\n";
    if (lm) definitions += "#define _LIGHTMAP_\n";
    if (lig) definitions += "#define _LIGHTING_\n";
    if (shad) definitions += "#define _SHADOWS_\n";
    if (morph) definitions += "#define _ANIMABLE_\n";
    if (skel) definitions += "#define _SKELETAL_\n";
    m_builtInPrograms[fog][tex][norm][lm][lig][shad][morph][skel] =
            createProgram(definitions + "\n" + source);
    return m_builtInPrograms[fog][tex][norm][lm][lig][shad][morph][skel];
}

m3iTexture *m3cRender::createTexture(uint16_t width, uint16_t height, const void *data, uint64_t size)
{
    return new m3cTexture(this, width, height, data, size);
}

m3cTextureRenderTarget *m3cRender::createTextureRenderTarget(m3cTexture *texture)
{
    return new m3cTextureRenderTarget(this, texture);
}

m3cProgram *m3cRender::createProgram(const std::string &source)
{
    return new m3cProgram(this, source);
}

void m3cRender::renderMesh(const m3sDrawMeshData &data, m3cProgram *program)
{
    bool first = true;
    for (auto &s : data.mesh->getData().surfaces)
    {
        renderMeshSurface(first, data, s, (program != nullptr ? program : getProgramForSurface(data, s)));
        first = false;
    }
}

void m3cRender::renderMeshSurface(const bool first, const m3sDrawMeshData &data, const m3sMeshSurface &surface, m3cProgram *program)
{
    m3cAbstractMesh *mesh = static_cast<m3cAbstractMesh*>(data.mesh);
    auto handles = program->getProgramHandles();
    auto mdata = mesh->getData();

    const uint8_t *dataPtr = reinterpret_cast<const uint8_t*>(mesh->getVertexPointer());
    auto firstFrameOffset = sizeof(m3sVertex) * int(data.frameTime) * mdata.verticesCount / mdata.framesCount;

    if (!mdata.hasMorphAnim)
        firstFrameOffset = 0;

    if (m_lastUsedProgram != program || first)
    {
        auto pos = glm::vec3(data.modelMatrix[3]);
        if (data.frustumEnabled && !m_frustum.cubeInFrustum(pos.x, pos.y, pos.z, data.mesh->getBoundingCubeSize()))
            return;

        if (data.depth)
            glEnable(GL_DEPTH_TEST);
        else
            glDisable(GL_DEPTH_TEST);

        m_shadowMapTexture->bind(7);

        program->use();
        mesh->bind();

        glVertexAttribPointer(handles.AttributePosition, 3, GL_FLOAT, GL_FALSE, sizeof(m3sVertex), reinterpret_cast<const uint8_t*>(dataPtr + 0 + firstFrameOffset));
        glVertexAttribPointer(handles.AttributeNormal, 3, GL_FLOAT, GL_FALSE, sizeof(m3sVertex), reinterpret_cast<const uint8_t*>(dataPtr + 12 + firstFrameOffset));
        glVertexAttribPointer(handles.AttributeTexcoord1, 2, GL_FLOAT, GL_FALSE, sizeof(m3sVertex), reinterpret_cast<const uint8_t*>(dataPtr + 24 + firstFrameOffset));
        glVertexAttribPointer(handles.AttributeColor, 4, GL_FLOAT, GL_FALSE, sizeof(m3sVertex), reinterpret_cast<const uint8_t*>(dataPtr + 32 + firstFrameOffset));

        if (m_frameData.fog.enabled)
        {
            glUniform1f(handles.UniformFogStart, m_frameData.fog.start);
            glUniform1f(handles.UniformFogEnd, m_frameData.fog.end);
            glUniform4fv(handles.UniformFogColor, 1, &m_frameData.fog.color[0]);
        }

        if (mdata.hasMorphAnim)
        {
            glEnableVertexAttribArray(handles.AttributePositionNext);
            glEnableVertexAttribArray(handles.AttributeNormalNext);

            auto nextFrame = data.frameTime + 1;
            if (nextFrame >= mdata.framesCount)
                nextFrame = data.frameTime;

            auto nextFrameOffset = sizeof(m3sVertex) * int(nextFrame) * data.mesh->getData().verticesCount / data.mesh->getData().framesCount;
            glVertexAttribPointer(handles.AttributePositionNext, 3, GL_FLOAT, GL_FALSE, sizeof(m3sVertex), reinterpret_cast<const uint8_t*>(dataPtr + 0 + nextFrameOffset));
            glVertexAttribPointer(handles.AttributeNormalNext, 3, GL_FLOAT, GL_FALSE, sizeof(m3sVertex), reinterpret_cast<const uint8_t*>(dataPtr + 12 + nextFrameOffset));

            glUniform1f(handles.UniformFrameTime, data.frameTime - int(data.frameTime));
        }
        else
        {
            glDisableVertexAttribArray(handles.AttributePositionNext);
            glDisableVertexAttribArray(handles.AttributeNormalNext);
        }

        if (mdata.hasSkeletalAnim)
        {
            glEnableVertexAttribArray(handles.AttributeJoint);
            glEnableVertexAttribArray(handles.AttributeWeight);

            glVertexAttribPointer(handles.AttributeJoint, 4, GL_FLOAT, GL_FALSE, sizeof(m3sVertex), reinterpret_cast<const uint8_t*>(dataPtr + 48));
            glVertexAttribPointer(handles.AttributeWeight, 4, GL_FLOAT, GL_FALSE, sizeof(m3sVertex), reinterpret_cast<const uint8_t*>(dataPtr + 64));

            glUniform1f(handles.UniformFrameTime, data.frameTime - int(data.frameTime));
            glUniformMatrix4fv(handles.UniformJointMatrix, mdata.skeleton->getJointsCount(), GL_FALSE, (const float*)mdata.skeleton->getJoints(data.frameTime).data());
        }
        else
        {
            glDisableVertexAttribArray(handles.AttributeJoint);
            glDisableVertexAttribArray(handles.AttributeWeight);
        }

        glUniform3fv(handles.UniformCameraPosition, 1, glm::value_ptr(m_frameData.cameraPosition));

        glUniformMatrix3fv(handles.UniformNormalMatrix, 1, GL_FALSE, glm::value_ptr(glm::inverseTranspose(glm::mat3(data.modelMatrix))));
        glUniformMatrix4fv(handles.UniformMvp, 1, GL_FALSE, glm::value_ptr(m_pvMatrix * data.modelMatrix));
        glUniformMatrix4fv(handles.UniformMv, 1, GL_FALSE, glm::value_ptr(m_frameData.viewMatrix * data.modelMatrix));
        glUniformMatrix4fv(handles.UniformM, 1, GL_FALSE, glm::value_ptr(data.modelMatrix));

        if (m_frameData.shadowsData.enabled)
        {
            glUniformMatrix4fv(handles.UniformShadowMapBiasMatrix, 1, GL_FALSE, glm::value_ptr(m_pvShadowsBiasMatrix * data.modelMatrix));
            glUniform1i(handles.UniformShadowMapTexture, 7);
        }

        if (m_frameData.lighting)
        {
            glUniform1f(handles.UniformLightsCount, glm::min(uint8_t(m3cProgram::getMaxLights()), m_frameData.usedLightsCount));
            glUniform4fv(handles.UniformSunColor, 1, glm::value_ptr(m_frameData.sun.color));
            glUniform3fv(handles.UniformSunPosition, 1, glm::value_ptr(m_frameData.sun.position));
            glUniform3fv(handles.UniformSunDirection, 1, glm::value_ptr(m_frameData.sun.direction));

            for (int i = 0; i < glm::min(m_frameData.usedLightsCount, m3cProgram::getMaxLights()); ++i)
            {
                glUniform4fv(handles.UniformLightDiffuse[i], 1, glm::value_ptr(m_frameData.lights[i].diffuseColor));
                glUniform3fv(handles.UniformLightPosition[i], 1, glm::value_ptr(m_frameData.lights[i].position));
                glUniform3fv(handles.UniformLightDirection[i], 1, glm::value_ptr(m_frameData.lights[i].direction));
                glUniform3fv(handles.UniformLightAttenuation[i], 1, glm::value_ptr(m_frameData.lights[i].attenuation));
            }
        }

        glUniform2f(handles.UniformScreenSize, getViewport().z, getViewport().w);

        if (data.blending)
        {
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
            glEnable(GL_BLEND);
        }
        else
        {
            glDisable(GL_BLEND);
        }

        if (data.depthMask)
        {
            glDepthFunc(GL_LESS);
            glDepthMask(GL_TRUE);
        }
        else
        {
            glDepthFunc(GL_LEQUAL);
            glDepthMask(GL_FALSE);
        }
    }//last used

    m_lastUsedProgram = program;

    if (surface.material.normalmapTexture)
    {
        glEnableVertexAttribArray(handles.AttributeTangent);
        glEnableVertexAttribArray(handles.AttributeBitangent);
        glVertexAttribPointer(handles.AttributeTangent, 3, GL_FLOAT, GL_FALSE, sizeof(m3sVertex), reinterpret_cast<const uint8_t*>(dataPtr + 80 + firstFrameOffset));
        glVertexAttribPointer(handles.AttributeBitangent, 3, GL_FLOAT, GL_FALSE, sizeof(m3sVertex), reinterpret_cast<const uint8_t*>(dataPtr + 92 + firstFrameOffset));
        static_cast<m3cTexture*>(surface.material.normalmapTexture)->bind(1);
        glUniform1i(handles.UniformNormalMapTexture, 1);
    }
    else
    {
        glDisableVertexAttribArray(handles.AttributeTangent);
        glDisableVertexAttribArray(handles.AttributeBitangent);
    }

    if (surface.material.lightmapTexture)
    {
        glEnableVertexAttribArray(handles.AttributeTexcoord2);
        glVertexAttribPointer(handles.AttributeTexcoord2, 2, GL_FLOAT, GL_FALSE, sizeof(m3sVertex), reinterpret_cast<const uint8_t*>(dataPtr + 104 + firstFrameOffset));
        static_cast<m3cTexture*>(surface.material.lightmapTexture)->bind(2);
        glUniform1i(handles.UniformLightMapTexture, 2);
    }
    else
    {
        glDisableVertexAttribArray(handles.AttributeTexcoord2);
    }

    if (surface.material.diffuseTexture)
    {
        static_cast<m3cTexture*>(surface.material.diffuseTexture)->bind(0);
        glUniform1i(handles.UniformTexture, 0);
    }

    glUniform4fv(handles.UniformMaterialDiffuse, 1, glm::value_ptr(surface.material.diffuseColor));
    glUniform4fv(handles.UniformMaterialSpecular, 1, glm::value_ptr(surface.material.specularColor));

    const uint8_t *elemDataPtr = reinterpret_cast<const uint8_t*>(mesh->getElementPointer());

    if (mdata.useElements)
        glDrawElements(GL_TRIANGLES, surface.count != 0 ? surface.count : mdata.elementsCount, GL_UNSIGNED_SHORT,
                       reinterpret_cast<const uint8_t*>(elemDataPtr + surface.offset * 2));
    else
        glDrawArrays(GL_TRIANGLES, surface.offset, surface.count != 0 ? surface.count : mdata.verticesCount);
}

m3cProgram *m3cRender::getProgramForSurface(const m3sDrawMeshData &data, const m3sMeshSurface &surface)
{
    if (data.is2d)
        return m_builtIn2dProgram;
    m3cAbstractMesh *mesh = static_cast<m3cAbstractMesh*>(data.mesh);
    return getBuiltInProgram(m_frameData.fog.enabled, surface.material.diffuseTexture != 0, surface.material.normalmapTexture != 0,
                             surface.material.lightmapTexture != 0, m_frameData.lighting && data.lightEnabled,
                             m_frameData.shadowsData.enabled && data.receiveShadows, mesh->getData().hasMorphAnim,
                             mesh->getData().hasSkeletalAnim);
}

uint64_t m3cRender::attachPostEffect(const std::string &fragmentSource, uint16_t width, uint16_t height)
{
    PostEffect effect;
    effect.width = width;
    effect.height = height;

    effect.tex = reinterpret_cast<m3cTexture*>(createTexture(effect.width, effect.height, 0, 0));
    effect.rt = createTextureRenderTarget(effect.tex);
    effect.prog = createProgram(
                "uniform sampler2D TEXTURE;                                                     \n"
                "uniform sampler2D SCREEN_TEXTURE;                                              \n"
                "uniform vec2 SCREEN_SIZE;                                                      \n"
                "varying vec2 texcoord;                                                         \n"
                "                                                                               \n"
                "#ifdef _VERTEX_                                                                \n"
                "                                                                               \n"
                "attribute vec3 POSITION;                                                       \n"
                "attribute vec2 TEXCOORD1;                                                       \n"
                "                                                                               \n"
                "void main()                                                                    \n"
                "{                                                                              \n"
                "    gl_Position = vec4(POSITION, 1.0);                                         \n"
                "    texcoord = TEXCOORD1;                                                       \n"
                "}                                                                              \n"
                "                                                                               \n"
                "#endif                                                                         \n"
                "                                                                               \n"
                "#ifdef _FRAGMENT_                                                              \n"
                "                                                                               \n"
                "void main()                                                                    \n"
                "{                                                                              \n"
                + fragmentSource +
                "}                                                                              \n"
                "                                                                               \n"
                "#endif                                                                         \n"

    );
    m_postEffectSets[m_currentPostEffectSet].postEffects.push_back(effect);
    return m_postEffectSets[m_currentPostEffectSet].postEffects.size() - 1;
}

void m3cRender::detachPostEffect(uint64_t id)
{
    delete m_postEffectSets[m_currentPostEffectSet].postEffects[id].tex;
    delete m_postEffectSets[m_currentPostEffectSet].postEffects[id].rt;
    delete m_postEffectSets[m_currentPostEffectSet].postEffects[id].prog;
}

void m3cRender::registerResource(m3iResource *res)
{
    m_registeredResources.push_back(res);
}

void m3cRender::unregisterResource(m3iResource *res)
{
    if (!m_dontUnregisterHandles)
        remove_from_vector(m_registeredResources, res);
}

void m3cRender::reload()
{
    m_lastUsedProgram = 0;

    for (auto r : m_registeredResources)
    {
        r->freeHandles();
    }

    for (auto r : m_registeredResources)
    {
        r->reloadHandles();
    }
}

}
