#pragma once

#include "m3hStd.h"
#include "m3iRender.h"
#include "m3cFrustum.h"
#include "m3sMeshSurface.h"
#include "m3sDrawMeshData.h"
#include "m3sRenderFrameData.h"
#include "m3iResourceManager.h"

namespace melissa
{

class m3cAbstractMesh;
class m3cTexture;
class m3cProgram;
class m3cTextureRenderTarget;

class m3cRender : public virtual m3iRender, public virtual m3iResourceManager
{
private:
    struct PostEffect
    {
        m3cTexture *tex = 0;
        m3cTextureRenderTarget *rt = 0;
        m3cProgram *prog = 0;
        uint16_t width = 0;
        uint16_t height = 0;
    };

    struct PostEffectSet
    {
        std::vector<PostEffect> postEffects;
    };

public:
    m3cRender();
    ~m3cRender();

    m3iMesh *createHardwareMesh(const m3sMeshData &data, bool optimize);
    m3iMesh *createLinkedMesh(const m3sMeshData &data);
    m3iTexture *createTexture(uint16_t width, uint16_t height, const void *data, uint64_t size);
    void setFrameData(const m3sRenderFrameData &data);
    void clear();
    void drawMesh(const m3sDrawMeshData &data);
    void flush();

    uint64_t createPostEffectSet();
    void clearPostEffectSet(uint64_t id);
    uint64_t getCurrentPostEffectSet() const;
    void usePostEffectSet(uint64_t id);

    uint64_t attachPostEffect(const std::string &fragmentSource, uint16_t width, uint16_t height);
    void detachPostEffect(uint64_t id);

    void registerResource(m3iResource *res);
    void unregisterResource(m3iResource *res);

    void reload();

    glm::vec4 getViewport() const;

private:
    m3cProgram *getBuiltInProgram(bool fog, bool tex, bool norm, bool lm, bool lig, bool shad, bool morph, bool skel);
    void renderScene();
    void burnShadowMap();
    m3cTextureRenderTarget *createTextureRenderTarget(m3cTexture *texture);
    m3cProgram *createProgram(const std::string &source);
    void renderMesh(const m3sDrawMeshData &data, m3cProgram *program = nullptr);
    void renderMeshSurface(const bool first, const m3sDrawMeshData &data, const m3sMeshSurface &surface, m3cProgram *program);
    m3cProgram *getProgramForSurface(const m3sDrawMeshData &data, const m3sMeshSurface &surface);

private:
    m3sRenderFrameData m_frameData;
    std::vector<m3iResource*> m_registeredResources;

    // draw frame data
    const glm::mat4 m_biasMatrix = glm::mat4(
                0.5, 0.0, 0.0, 0.0,
                0.0, 0.5, 0.0, 0.0,
                0.0, 0.0, 0.5, 0.0,
                0.5, 0.5, 0.5, 1.0
    );
    glm::mat4 m_pvMatrix, m_pvShadowsBiasMatrix;
    //! [fog][texturing][normalmapping][lighting][shadows][animable][skel]
    m3cProgram *m_builtInPrograms[2][2][2][2][2][2][2][2];
    m3cProgram *m_builtIn2dProgram;
    std::vector<m3sDrawMeshData> m_meshesForDraw;
    //! [skeletal]
    m3cProgram *m_depthProgram[2] = {0, 0};
    m3cTexture *m_shadowMapTexture = 0;
    m3cTextureRenderTarget *m_shadowsRenderTarget = 0;
    bool m_dontUnregisterHandles = false;

    m3cFrustum m_frustum;
    m3cAbstractMesh *m_postPlaneMesh = 0;
    PostEffect m_clearEffect;
    std::vector<PostEffectSet> m_postEffectSets;
    uint64_t m_currentPostEffectSet = 0;

    const m3cProgram *m_lastUsedProgram = 0;
};

}
