#include "m3cBatch.h"
#include "m3iRender.h"
#include "m3sMeshData.h"
#include "m3cAbstractMesh.h"
#include "m3sDrawMeshData.h"
#include "m3iBatchDrawDelegate.h"

namespace
{

melissa::m3sVertex vert(const glm::vec3 &pos, const glm::vec2 &tc, const glm::vec4 &color)
{
    melissa::m3sVertex v;
    v.position = pos;
    v.color = color;
    v.texcoord1 = tc;
    return v;
}

}

namespace melissa
{

m3cBatch::m3cBatch(m3iMovableManager *movMan, m3iRender *render, const glm::vec2 &spaceSize, m3iTexture *texture, m3iBatchDrawDelegate *drawDel, uint64_t max, bool is2d)
    : m3cMovable(movMan),
      m_max(max),
      m_texture(texture),
      m_drawDel(drawDel),
      m_vertices(m_max * 6),
      m_spaceSize(spaceSize)
{
    assert(max > 0);

    m3sMeshSurface surface;
    surface.material.diffuseTexture = texture;

    m3sMeshData data;
    data.useElements = false;
    data.vertices = m_vertices.data();
    data.verticesCount = m_vertices.size();
    data.surfaces.push_back(surface);

    m_mesh = render->createLinkedMesh(data);
    m_drawMeshData.is2d = is2d;
    m_drawMeshData.blending = true;
    m_drawMeshData.castShadows = false;
    m_drawMeshData.depth = !is2d;
    m_drawMeshData.frustumEnabled = false;
    m_drawMeshData.lightEnabled = false;
    m_drawMeshData.mesh = m_mesh;
    m_drawMeshData.receiveShadows = false;
}

void m3cBatch::draw(m3iRender *render, m3sRenderFrameData &frameData)
{
    if (!isVisible())
        return;

    m_batchDraws.clear();
    m_drawDel->onDraw(this);

    for (uint64_t i = 0; i < m_batchDraws.size(); ++i)
    {
        addBatchDrawData(i, m_batchDraws[i]);
    }

    static_cast<m3cAbstractMesh*>(m_mesh)->editSurfaces()[0].count = m_batchDraws.size() * 6;
    m_drawMeshData.modelMatrix = getModelMatrix(true);

    render->drawMesh(m_drawMeshData);
}

void m3cBatch::draw(const m3sBatchDrawData &data)
{
    if (!isVisible())
        return;

    m_batchDraws.push_back(data);
}

glm::vec4 m3cBatch::convertFrame(const glm::vec4 &frame)
{
    auto result = frame;
    float sz = glm::abs(result.w - result.y);
    result.y = 1 - result.w;
    result.w = 1 - result.w + sz;
    return result;
}

void m3cBatch::addBatchDrawData(uint64_t index, const m3sBatchDrawData &data)
{
    glm::mat4 model =
            glm::translate(data.position)
            * glm::rotate(glm::radians(data.angle), glm::vec3(0, 0, 1))
            * glm::scale(glm::vec3(data.size.x, data.size.y, 1.0));

    if (m_drawMeshData.is2d)
        model = glm::ortho(0.0f, m_spaceSize.x, m_spaceSize.y, 0.0f, -10.0f, 10.0f) * model;

    auto tc = convertFrame(data.texcoordSquare);

    m3sVertex quad[4] = {
        vert(glm::vec3(0 - data.anchor.x, 0 - data.anchor.y, 0), glm::vec2(tc[0], tc[3]), data.color),
        vert(glm::vec3(0 - data.anchor.x, 1 - data.anchor.y, 0), glm::vec2(tc[0], tc[1]), data.color),
        vert(glm::vec3(1 - data.anchor.x, 1 - data.anchor.y, 0), glm::vec2(tc[2], tc[1]), data.color),
        vert(glm::vec3(1 - data.anchor.x, 0 - data.anchor.y, 0), glm::vec2(tc[2], tc[3]), data.color)
    };

    for (int i = 0; i < 4; ++i)
        quad[i].position = glm::vec3(model * glm::vec4(quad[i].position, 1));

    m_vertices[index * 6 + 0] = quad[0];
    m_vertices[index * 6 + 1] = quad[1];
    m_vertices[index * 6 + 2] = quad[2];
    m_vertices[index * 6 + 3] = quad[0];
    m_vertices[index * 6 + 4] = quad[2];
    m_vertices[index * 6 + 5] = quad[3];
}

}
