#include <sstream>
#include "m3hGlm.h"
#include "m3iFile.h"
#include "m3cScene.h"
#include "m3iModel.h"
#include "m3cLogger.h"
#include "m3iRender.h"
#include "m3sMeshData.h"
#include "m3cSkeleton.h"
#include "m3iFileSystem.h"
#include "m3sSkeletonData.h"
#include "m3cModelSetsLoader.h"

namespace
{

void calculateTangent(melissa::m3sVertex &v0, melissa::m3sVertex &v1, melissa::m3sVertex &v2)
{
    glm::vec3 deltaPos1 = v1.position - v0.position;
    glm::vec3 deltaPos2 = v2.position - v0.position;
    glm::vec2 deltaUV1 = v1.texcoord1 - v0.texcoord1;
    glm::vec2 deltaUV2 = v2.texcoord1 - v0.texcoord1;
    float r = 1.0f / (deltaUV1.x * deltaUV2.y - deltaUV1.y * deltaUV2.x);
    glm::vec3 tangent = (deltaPos1 * deltaUV2.y   - deltaPos2 * deltaUV1.y)*r;
    glm::vec3 bitangent = (deltaPos2 * deltaUV1.x   - deltaPos1 * deltaUV2.x)*r;
    v0.tangent = tangent;
    v0.bitangent = bitangent;
    v1.tangent = tangent;
    v1.bitangent = bitangent;
    v2.tangent = tangent;
    v2.bitangent = bitangent;
}

template <typename T>
T lex_cast(const std::string &str)
{
    T t;
    std::stringstream ss(str);
    ss >> t;
    return t;
}

std::string readLine(melissa::m3iFile *file)
{
    char c = 0;
    std::string res;

    while (true)
    {
        if (file->isEof())
            break;
        file->read(&c, 1);
        if (c == '\n')
            break;
        if (c != '\r')
            res += c;
    }

    return res;
}

std::vector<std::string> &split(const std::string &s, char delim, std::vector<std::string> &elems)
{
    std::stringstream ss(s);
    std::string item;
    while (std::getline(ss, item, delim)) {
        elems.push_back(item);
    }
    return elems;
}


std::vector<std::string> split(const std::string &s, char delim)
{
    std::vector<std::string> elems;
    split(s, delim, elems);
    return elems;
}

struct Material
{
    std::string name;
    std::string diffuseTexture, normalTexture;
    glm::vec4 materialDiffuse = {1, 1, 1, 1};
    glm::vec4 materialSpecular = {0, 0, 0, 0};
};

struct Vertex
{
    float x, y, z;
};

struct Normal
{
    float x, y, z;
};

struct Texcoord
{
    float u1, v1;
    float u2, v2;
};

}

namespace melissa
{

std::vector<m3sModelSet> m3cModelSetsLoader::loadModelSets(m3cScene *scene, m3iRender *render, m3iFileSystem *filesystem, const std::string &filename)
{
    auto it = m_cachedScenes.find(filename);
    if (it != m_cachedScenes.end())
        return createModelSets(it->second, scene);

    if (filename.substr(filename.find_last_of('.') + 1) != "m3json")
        return std::vector<m3sModelSet>();

    auto f = filesystem->openFile(filename);
    std::vector<char> str(f->getSize());
    f->read(str.data(), f->getSize());
    delete f;
    f = 0;

    Json::Value root;
    Json::Reader reader;
    bool parsingSuccessful = reader.parse(str.data(), root);
    if (!parsingSuccessful)
    {
        m3cLogger::fatal()  << "Failed to parse json mesh"
                   << reader.getFormattedErrorMessages();
        return std::vector<m3sModelSet>();
    }

    CachedScene cachedScene;
    const Json::Value objects = root["objects"];
    for (int i = 0; i < objects.size(); ++i)
    {
        createObject(cachedScene, objects[i], scene, render);
    }

    m_cachedScenes[filename] = cachedScene;
    return createModelSets(cachedScene, scene);
}

std::vector<m3sModelSet> m3cModelSetsLoader::createModelSets(const m3cModelSetsLoader::CachedScene &cachedScene, melissa::m3cScene *scene)
{
    std::vector<m3sModelSet> result;
    for (const CachedObject &o : cachedScene.objects)
    {
        m3iModel *model = scene->createModel(o.mesh);
        model->setPosition(o.pos);
        model->setQuaternion(o.ori);
        model->setScale(o.scale);
        m3sModelSet set;
        set.model = model;
        set.name = o.name;
        result.push_back(set);
    }
    return result;
}

void m3cModelSetsLoader::createObject(m3cModelSetsLoader::CachedScene &cachedScene, const Json::Value &object, m3cScene *scene, m3iRender *render)
{
    std::vector<m3sVertex> vertices;
    const Json::Value vertex = object["vertex"];
    const int numvals = 18;
    for (int index = 0; index < vertex.size() / numvals; ++index)
    {
        m3sVertex v;
        v.position.x = vertex[index * numvals + 0].asFloat();
        v.position.y = vertex[index * numvals + 1].asFloat();
        v.position.z = vertex[index * numvals + 2].asFloat();
        v.normal.x = vertex[index * numvals + 3].asFloat();
        v.normal.y = vertex[index * numvals + 4].asFloat();
        v.normal.z = vertex[index * numvals + 5].asFloat();
        v.texcoord1.x = vertex[index * numvals + 6].asFloat();
        v.texcoord1.y = vertex[index * numvals + 7].asFloat();
        v.texcoord2.x = vertex[index * numvals + 8].asFloat();
        v.texcoord2.y = vertex[index * numvals + 9].asFloat();
        v.joint.x = vertex[index * numvals + 10].asInt();
        v.joint.y = vertex[index * numvals + 11].asInt();
        v.joint.z = vertex[index * numvals + 12].asInt();
        v.joint.w = vertex[index * numvals + 13].asInt();
        v.weight.x = vertex[index * numvals + 14].asFloat();
        v.weight.y = vertex[index * numvals + 15].asFloat();
        v.weight.z = vertex[index * numvals + 16].asFloat();
        v.weight.w = vertex[index * numvals + 17].asFloat();
        vertices.push_back(v);
    }

    std::vector<m3sMeshSurface> surfaces;
    const Json::Value surface = object["surface"];
    for (int index = 0; index < surface.size(); ++index)
    {
        m3sMeshSurface s;
        s.material.name = surface[index]["name"].asString();
        s.offset = surface[index]["start"].asInt64();
        s.count = surface[index]["count"].asInt64();
        s.material.diffuseColor.r = surface[index]["diffuseColor"][0].asFloat();
        s.material.diffuseColor.g = surface[index]["diffuseColor"][1].asFloat();
        s.material.diffuseColor.b = surface[index]["diffuseColor"][2].asFloat();
        s.material.diffuseColor.a = surface[index]["diffuseColor"][3].asFloat();
        s.material.diffuseTexture = scene->loadTexture(surface[index]["diffuseTexture"].asString());
        s.material.normalmapTexture = scene->loadTexture(surface[index]["normalmapTexture"].asString());
        s.material.lightmapTexture = scene->loadTexture(surface[index]["lightmapTexture"].asString());
        surfaces.push_back(s);
    }

    std::vector<int64_t> parentIndices;
    m3sSkeletonData skelData;
    const Json::Value skeleton = object["skeleton"];
    for (int index = 0; index < skeleton.size(); ++index)
    {
        const Json::Value joint = skeleton[index];
        m3sSkeletonData::Joint j;
        j.pos[0] = joint["pose"][0].asFloat();
        j.pos[1] = joint["pose"][1].asFloat();
        j.pos[2] = joint["pose"][2].asFloat();
        j.quat[0] = joint["pose"][3].asFloat();
        j.quat[1] = joint["pose"][4].asFloat();
        j.quat[2] = joint["pose"][5].asFloat();
        j.quat[3] = joint["pose"][6].asFloat();
        parentIndices.push_back(joint["parentIndex"].asInt64());
        const Json::Value jointFrame = joint["frame"];
        for (int indexFrame = 0; indexFrame < jointFrame.size() / 7; ++indexFrame)
        {
            m3sSkeletonData::Frame jf;
            jf.time = indexFrame;
            jf.pos[0] = jointFrame[indexFrame * 7 + 0].asFloat();
            jf.pos[1] = jointFrame[indexFrame * 7 + 1].asFloat();
            jf.pos[2] = jointFrame[indexFrame * 7 + 2].asFloat();
            jf.quat[0] = jointFrame[indexFrame * 7 + 3].asFloat();
            jf.quat[1] = jointFrame[indexFrame * 7 + 4].asFloat();
            jf.quat[2] = jointFrame[indexFrame * 7 + 5].asFloat();
            jf.quat[3] = jointFrame[indexFrame * 7 + 6].asFloat();
            j.frames.push_back(jf);
        }
        skelData.joints.push_back(j);
    }

    for (int64_t i = 0; i < parentIndices.size(); ++i)
    {
        if (parentIndices[i] >= 0)
        {
            skelData.joints[i].parentIdx = parentIndices[i];
        }
    }

    m3cSkeleton *cskel = new m3cSkeleton;
    cskel->build(skelData);

    for (uint64_t i = 0; i < vertices.size() / 3; ++i)
    {
        m3sVertex &v1 = vertices[i * 3 + 0];
        m3sVertex &v2 = vertices[i * 3 + 1];
        m3sVertex &v3 = vertices[i * 3 + 2];
        calculateTangent(v1, v2, v3);
    }

    m3sMeshData data;
    data.useElements = false;
    data.vertices = vertices.data();
    data.verticesCount = vertices.size();
    data.surfaces = surfaces;

    if (skeleton.size())
    {
        data.skeleton = cskel;
        data.hasSkeletalAnim = true;
    }

    CachedObject co;
    co.name = object["name"].asString();
    co.mesh = render->createHardwareMesh(data, false);
    co.pos[0] = object["transformations"][0].asFloat();
    co.pos[1] = object["transformations"][1].asFloat();
    co.pos[2] = object["transformations"][2].asFloat();
    co.ori.x = object["transformations"][3].asFloat();
    co.ori.y = object["transformations"][4].asFloat();
    co.ori.z = object["transformations"][5].asFloat();
    co.ori.w = object["transformations"][6].asFloat();
    co.scale.x = object["transformations"][7].asFloat();
    co.scale.y = object["transformations"][8].asFloat();
    co.scale.z = object["transformations"][9].asFloat();

    cachedScene.objects.push_back(co);
}

}
