#include "m3cMovable.h"
#include "m3iMovableManager.h"

namespace
{

glm::vec3 localXAxis(const glm::vec3 &direction){ return glm::cross(direction, glm::vec3(0, -1, 0)); }
glm::vec3 localYAxis(const glm::vec3 &direction){ return glm::cross(direction, localXAxis(direction)); }

glm::mat4 objLookAt(const glm::vec3 &topos, const glm::vec3 &pos, const glm::vec3 &dir){
    glm::mat4 m=glm::inverse(glm::lookAt(glm::vec3(0, 0, 0), topos-pos, localYAxis(dir)));
    return m;
}

}

namespace melissa
{

m3cMovable::m3cMovable(m3iMovableManager *movMan)
    : m_movMan(movMan)
{
    m_movMan->registerMovable(this);
}

m3cMovable::~m3cMovable()
{
    m_movMan->unregisterMovable(this);
}

void m3cMovable::addPosition(const glm::vec3 &position)
{
    m_translation += position;
    m_modify = true;
}

void m3cMovable::setPosition(const glm::vec3 &position)
{
    m_translation = position;
    m_modify = true;
}

void m3cMovable::setEuler(const glm::vec3 &rotation)
{
    m_orientation = glm::quat(glm::radians(rotation));
    m_modify = true;
}

void m3cMovable::setQuaternion(const glm::quat &quaternion)
{
    m_orientation = quaternion;
    m_modify = true;
}

void m3cMovable::setScale(const glm::vec3 &scale)
{
    m_scale = scale;
    m_modify = true;
}

void m3cMovable::pitch(float v, bool global)
{
    rotate(glm::toQuat(glm::yawPitchRoll(0.0f, glm::radians(v), 0.0f)), global);
}

void m3cMovable::yaw(float v, bool global)
{
    rotate(glm::toQuat(glm::yawPitchRoll(glm::radians(v), 0.0f, 0.0f)), global);
}

void m3cMovable::roll(float v, bool global)
{
    rotate(glm::toQuat(glm::yawPitchRoll(0.0f, 0.0f, glm::radians(v))), global);
}

void m3cMovable::rotate(const glm::quat &orientation, bool global)
{
    if (!global)
        m_orientation *= orientation;
    else
        m_orientation = orientation * m_orientation;
    m_modify = true;
}

void m3cMovable::addScale(const glm::vec3 &scale)
{
    m_scale *= scale;
    m_modify = true;
}

glm::vec3 m3cMovable::getPosition(bool global) const
{
    return glm::vec3(getModelMatrix(global)[3]);
}

glm::vec3 m3cMovable::getDirection(bool global) const
{
    return glm::vec3(getModelMatrix(global)[2]);
}

glm::vec3 m3cMovable::getEuler(bool global) const
{
    return glm::degrees(glm::eulerAngles(glm::quat(getModelMatrix(global))));
}

glm::vec3 m3cMovable::getScale(bool global) const
{
    return getScaleFromMatrix(getModelMatrix(global));
}

glm::mat4 m3cMovable::getModelMatrix(bool global) const
{
    if (m_modify)
    {
        m_mat = glm::translate(m_translation)
                * glm::toMat4(m_orientation)
                * glm::scale(m_scale);
        m_modify = false;
    }

    if (!global || !m_parent)
        return m_mat;

    return m_parent->getModelMatrix(true) * m_mat;
}

void m3cMovable::setModelMatrix(const glm::mat4 &matrix)
{
    m_modify = false;
    auto p = glm::vec3(matrix[3]);
    auto r = glm::eulerAngles(glm::quat(matrix));
    r *= 180.0f / 3.1415f;
// TODO: scale from matrix
    setPosition(p);
    setEuler(r);
    setScale(getScaleFromMatrix(matrix));
}

void m3cMovable::setParent(const m3iMovable *parent)
{
    m_parent = parent;
}

void m3cMovable::lookTo(const glm::vec3 &position)
{
    auto m = getModelMatrix(true);
    m_orientation = glm::toQuat(objLookAt(position, glm::vec3(m[3]), glm::vec3(m[2])));
}

glm::vec3 m3cMovable::getScaleFromMatrix(const glm::mat4 &mat) const
{
    glm::vec3 col1(mat[0][0], mat[0][1], mat[0][2]);
    glm::vec3 col2(mat[1][0], mat[1][1], mat[1][2]);
    glm::vec3 col3(mat[2][0], mat[2][1], mat[2][2]);

    glm::vec3 scaling;
    scaling.x = glm::length(col1);
    scaling.y = glm::length(col2);
    scaling.z = glm::length(col3);
    return scaling;
}

}
