#include "m3iRender.h"
#include "m3iRender.h"
#include "m3sMeshData.h"
#include "m3cParticles.h"
#include "m3sDrawMeshData.h"
#include "m3sRenderFrameData.h"
#include "m3iParticlesBehaviour.h"

namespace
{

melissa::m3sVertex vert(const glm::vec3 &pos, const glm::vec2 &tc, const glm::vec4 &color)
{
    melissa::m3sVertex v;
    v.position = pos;
    v.color = color;
    v.texcoord1 = tc;
    return v;
}

}

namespace melissa
{

m3cParticles::m3cParticles(m3iMovableManager *movMan, m3iRender *render, m3iTexture *texture, m3iParticlesBehaviour *behav, uint64_t units)
    : m3cMovable(movMan),
      m_units(units),
      m_texture(texture),
      m_behav(behav),
      m_vertices(m_units * 6),
      m_particleUnits(m_units)
{
    m3sMeshSurface surface;
    surface.material.diffuseTexture = texture;

    m3sMeshData mdata;
    mdata.vertices = m_vertices.data();
    mdata.verticesCount = m_vertices.size();
    mdata.elements = 0;
    mdata.elementsCount = 0;
    mdata.useElements = false;
    mdata.surfaces.push_back(surface);

    m_mesh = render->createLinkedMesh(mdata);

    if (m_behav)
        m_behav->initialize(m_particleUnits);
}

void m3cParticles::setForEachMatrix(bool forEachMatrix)
{
    m_forEachMatrix = forEachMatrix;
}

namespace
{

glm::vec3 localXAxis(const glm::vec3 &direction){ return glm::cross(direction, glm::vec3(0, -1, 0)); }
glm::vec3 localYAxis(const glm::vec3 &direction){ return glm::cross(direction, localXAxis(direction)); }

glm::mat4 objLookAt(const glm::vec3 &topos, const glm::vec3 &pos, const glm::vec3 &dir){
    glm::mat4 m=glm::inverse(glm::lookAt(pos, topos-pos, localYAxis(dir)));
    return m;
}

}

void m3cParticles::draw(m3iRender *render, m3sRenderFrameData &frameData)
{
    for (uint64_t i = 0; i < m_particleUnits.size(); ++i)
        setParticleUnitToVertices(frameData, i, m_particleUnits[i]);

    if (!isVisible())
        return;

    m3sDrawMeshData data;
    data.lightEnabled = false;
    data.depthMask = false;
    data.blending = true;
    data.depth = true;
    data.castShadows = false;
    data.receiveShadows = false;
    data.frustumEnabled = false;
    data.mesh = m_mesh;

    if (m_forEachMatrix)
    {
        data.modelMatrix = glm::mat4(1.0);
    }
    else
    {
        data.modelMatrix = getModelMatrix(true);
        glm::mat4 bmat = objLookAt(frameData.cameraPosition, glm::vec3(data.modelMatrix[3]), glm::vec3(data.modelMatrix[2]));
        data.modelMatrix = bmat;
    }

    render->drawMesh(data);
}

void m3cParticles::update(double dt)
{
    if (m_behav)
        m_behav->update(m_particleUnits, dt);
}

void m3cParticles::setState(m3eParticlesState state)
{
    if (m_behav)
        m_behav->changedState(state);
}

void m3cParticles::setParticleUnitToVertices(m3sRenderFrameData &frameData, uint64_t index, const m3sParticleUnit &unit)
{
    const m3sVertex vertices[6] = {
        vert(glm::vec3(-0.5, -0.5, 0), glm::vec2(0, 0), glm::vec4()),
        vert(glm::vec3(-0.5, 0.5, 0), glm::vec2(0, 1), glm::vec4()),
        vert(glm::vec3(0.5, 0.5, 0), glm::vec2(1, 1), glm::vec4()),

        vert(glm::vec3(-0.5, -0.5, 0), glm::vec2(0, 0), glm::vec4()),
        vert(glm::vec3(0.5, 0.5, 0), glm::vec2(1, 1), glm::vec4()),
        vert(glm::vec3(0.5, -0.5, 0), glm::vec2(1, 0), glm::vec4())
    };

    glm::mat4 unitmat;
    if (m_forEachMatrix)
    {
        auto scale = getScaleFromMatrix(unit.matrix);
        unitmat = getModelMatrix(true) * unit.matrix;
        unitmat = objLookAt(frameData.cameraPosition, glm::vec3(unitmat[3]), glm::vec3(unitmat[2])) * glm::scale(scale);
    }
    else
    {
        unitmat = unit.matrix;
    }

    for (uint8_t i = 0; i < 6; ++i)
    {
        glm::vec4 pos = unitmat * glm::vec4(vertices[i].position, 1.0);
        m_vertices[index * 6 + i].position.x = pos.x;
        m_vertices[index * 6 + i].position.y = pos.y;
        m_vertices[index * 6 + i].position.z = pos.z;
        m_vertices[index * 6 + i].texcoord1.x = vertices[i].texcoord1.x;
        m_vertices[index * 6 + i].texcoord1.y = vertices[i].texcoord1.y;
        m_vertices[index * 6 + i].texcoord2.x = vertices[i].texcoord2.x;
        m_vertices[index * 6 + i].texcoord2.y = vertices[i].texcoord2.y;
        m_vertices[index * 6 + i].color.r = unit.color.r;
        m_vertices[index * 6 + i].color.g = unit.color.g;
        m_vertices[index * 6 + i].color.b = unit.color.b;
        m_vertices[index * 6 + i].color.a = unit.color.a;
    }
}

}
