#include "m3cRoot.h"
#include "m3cScene.h"
#include "m3iSystem.h"
#include "m3cRender.h"
#include "m3iFileSystem.h"
#include "m3iApplication.h"
#include "m3iBatchContext.h"
#include "m3sBatchDrawData.h"
#include "m3iBatchDrawDelegate.h"

namespace
{

using namespace melissa;

class m3cRootApplication : public m3iApplication, public m3iBatchDrawDelegate
{
public:
    m3cRootApplication(m3iRootApplication *rootApp)
    {
        m_rootApplication = rootApp;
    }

    m3eScreenOrientation getSimulatorOrientation() const
    {
        return m_rootApplication->getSimulatorOrientation();
    }

    void initialize(m3iSystem *system)
    {
        m_system = system;
        static_cast<m3cRoot*>(getRoot())->initialize(system);
        m_defaultScene = getRoot()->getScene();
        m_loadingScene = getRoot()->createScene(true);
        if (!m_rootApplication->loadingScreenFilename().empty())
            m_loadingScene->createBatch(glm::vec2(1, 1), m_rootApplication->loadingScreenFilename(), this, 1, true);
    }

    void render()
    {
        if (m_loadingStage == 2)
        {
            getRoot()->setScene(m_defaultScene);
            if (m_needInitialize)
                m_rootApplication->initialize(m_system, getRoot());
            m_needInitialize = false;
        }

        if (m_loadingStage < 3)
            m_loadingStage++;

        static auto lastTime = m_system->getTime();
        auto time = m_system->getTime();
        auto dt = time - lastTime;
        lastTime = time;

        if (dt > 0.3)
            dt = 0.3;

        static_cast<m3cRoot*>(getRoot())->render();

        if (isLoaded())
            m_rootApplication->render(dt);
    }

    void onDraw(m3iBatchContext *ctx)
    {
        m3sBatchDrawData data;
        data.size = glm::vec2(1, 1);
        ctx->draw(data);
    }

    void reload()
    {
        if (!isLoaded())
            return;

        static_cast<m3cRoot*>(getRoot())->reload();
        m_loadingStage = 0;
        m_defaultScene = getRoot()->getScene();
        getRoot()->setScene(m_loadingScene);
        m_rootApplication->reload();
    }

    void touch(m3eInputAction act, uint8_t id, uint16_t x, uint16_t y)
    {
        if (isLoaded())
            m_rootApplication->touch(act, id, x, y);
    }

    void keyboard(m3eInputAction act, uint8_t key)
    {
        if (isLoaded())
            m_rootApplication->keyboard(act, key);
    }

    void stop()
    {
        if (isLoaded())
            m_rootApplication->stop();
    }

    void close()
    {
        if (isLoaded())
            m_rootApplication->close();
    }

    ~m3cRootApplication()
    {
        delete m_rootApplication;
    }

private:
    bool isLoaded() const
    {
        return m_loadingStage == 3;
    }

private:
    int m_loadingStage = 0;
    bool m_needInitialize = true;
    m3iSystem *m_system = 0;
    m3iScene *m_defaultScene = 0;
    m3iScene *m_loadingScene = 0;
    m3iRootApplication *m_rootApplication = 0;
};

}

melissa::m3iApplication *createApplication(m3iRootApplication *rootApp)
{
    return new m3cRootApplication(rootApp);
}

namespace melissa
{

m3iRoot *getRoot()
{
    static m3cRoot root(new m3cRender);
    return &root;
}

m3cRoot::m3cRoot(m3iRender *render)
    : m_render(render)
{
}

m3cRoot::~m3cRoot()
{
    m_dontUnregisterScenes = true;
    for (auto s: m_registeredScenes)
        delete s;
    delete m_render;
}

m3iScene *m3cRoot::createScene(bool setAsCurrent)
{
    auto scene = new m3cScene(m_render, m_system, this);
    if (setAsCurrent)
        setScene(scene);
    return scene;
}

void m3cRoot::setScene(m3iScene *scene)
{
    m_scene = static_cast<m3cScene*>(scene);
}

m3iScene *m3cRoot::getScene() const
{
    return m_scene;
}

m3iFile *m3cRoot::openFile(const std::string &filename)
{
    return m_system->getFileSystem()->openFile(filename);
}

void m3cRoot::registerScene(m3iScene *scene)
{
    m_registeredScenes.push_back(scene);
}

void m3cRoot::unregisterScene(m3iScene *scene)
{
    if (m_dontUnregisterScenes)
        return;
    remove_from_vector(m_registeredScenes, scene);
}

void m3cRoot::initialize(m3iSystem *system)
{
    m_system = system;
    createScene(true);
}

void m3cRoot::render()
{
    m_lastViewport = m_render->getViewport();

    static auto lastTime = m_system->getTime();
    auto time = m_system->getTime();
    auto dt = time - lastTime;
    lastTime = time;

    if (dt > 0.3)
        dt = 0.3;
    m_scene->render(dt);
}

void m3cRoot::reload()
{
    m_render->reload();
    if (m_scene)
        m_scene->reload();
}

glm::vec4 m3cRoot::getViewport() const
{
    return m_lastViewport;
}

}
