#include "m3iMesh.h"
#include "m3cScene.h"
#include "m3cModel.h"
#include "m3cLight.h"
#include "m3cCamera.h"
#include "m3iSystem.h"
#include "m3cBatch.h"
#include "m3iRender.h"
#include "m3cMovable.h"
#include "m3sMeshData.h"
#include "m3cParticles.h"
#include "m3cPhysicsBody.h"
#include "m3cPhysicsWorld.h"
#include "m3iSceneManager.h"
#include "m3cTextureLoader.h"
#include "m3sRenderFrameData.h"
#include "m3iBatchDrawDelegate.h"

namespace
{

using namespace melissa;

class LoadingDrawDelegate : public m3iBatchDrawDelegate
{
public:
    void onDraw(m3iBatchContext *ctx)
    {
        m3sBatchDrawData data;
        data.size = glm::vec2(1, 1);
        ctx->draw(data);
    }
};

}

namespace melissa
{

m3cScene::m3cScene(m3iRender *render, m3iSystem *system, m3iSceneManager *sceneMan)
    : m_render(render),
      m_system(system),
      m_sceneMan(sceneMan),
      m_physicsWorld(new m3cPhysicsWorld(this))
{
    m_sceneMan->registerScene(this);
    m_postEffectSet = m_render->createPostEffectSet();
    createCamera(true);
}

m3cScene::~m3cScene()
{
    m_dontUnregisterMovables = true;
    for (auto m : m_registeredMovables)
        delete m;
    m_sceneMan->unregisterScene(this);
}

void m3cScene::reload()
{
}

m3iBatch *m3cScene::createBatch(const glm::vec2 &spaceSize, const std::string &filename, m3iBatchDrawDelegate *drawDel, uint64_t max, bool is2d)
{
    auto b = new m3cBatch(this, m_render, spaceSize, loadTexture(filename), drawDel, max, is2d);
    m_batchesForDraw.push_back(b);
    return b;
}

m3iMovable *m3cScene::createMovable()
{
    return new m3cMovable(this);
}

std::vector<m3sModelSet> m3cScene::loadModelSets(const std::string &filename)
{
    return m_modelSetsLoader.loadModelSets(this, m_render, m_system->getFileSystem(), filename);
}

m3iModel *m3cScene::createModel(m3iMesh *mesh)
{
    auto model = new m3cModel(this, mesh);
    m_modelsForDraw.push_back(model);
    return model;
}

m3iCamera *m3cScene::createCamera(bool setAsCurrent)
{
    auto camera = new m3cCamera(this);
    if (setAsCurrent)
        setCamera(camera);
    return camera;
}

void m3cScene::setCamera(m3iCamera *camera)
{
    m_camera = static_cast<m3cCamera*>(camera);
}

m3iCamera *m3cScene::getCamera() const
{
    return m_camera;
}

m3iLight *m3cScene::createLight()
{
    auto light = new m3cLight(this);
    m_lightsForUse.push_back(light);
    return light;
}

m3iParticles *m3cScene::createParticles(m3iParticlesBehaviour *behav, const std::string &filename, uint64_t count)
{
    auto p = new m3cParticles(this, m_render, loadTexture(filename), behav, count);
    m_particlesForDraw.push_back(p);
    return p;
}

void m3cScene::render(double dt)
{
    m_physicsWorld->simulate(dt);
    for (auto it = m_attachedPhysicsBodies.begin(); it != m_attachedPhysicsBodies.end(); ++it)
    {
        it->first->setModelMatrix(it->second->getModelMatrix());
    }

    m_render->usePostEffectSet(m_postEffectSet);

    auto vp = m_render->getViewport();
    m_frameData.projectionMatrix = glm::perspective(glm::radians(60.0f), vp[2] / vp[3], 1.0f, 300.0f);
    m_frameData.cameraPosition = m_camera->getPosition(true);

    m_camera->use(m_render, m_frameData);

    m_frameData.usedLightsCount = m_lightsForUse.size();
    for (uint8_t i = 0; i < 4; ++i)
    {
        if (m_lightsForUse.size() > i)
        {
            m_lightsForUse[i]->use(i, m_render, m_frameData);
        }
        else
        {
            m_frameData.lights[i].diffuseColor = glm::vec4(0, 0, 0, 1);
        }
    }

    m_render->clear();
    m_render->setFrameData(m_frameData);

    auto mfd = m_modelsForDraw;
    std::vector<m3cModel*> bmfd;

    for (auto &m : mfd)
    {
        m->update(dt);
        if (m->isBlendingEnabled())
            bmfd.push_back(m);
        else
            m->draw(m_render, m_frameData);
    }

    std::sort(bmfd.begin(), bmfd.end(), [&](m3cModel *m1,m3cModel *m2){
        return glm::distance(m_camera->getPosition(true), m1->getPosition(true)) >
                glm::distance(m_camera->getPosition(true), m2->getPosition(true));
    });

    for (auto &m : bmfd)
    {
        m->draw(m_render, m_frameData);
    }

    for (auto b : m_batchesForDraw)
    {
        b->draw(m_render, m_frameData);
    }
    for (auto p : m_particlesForDraw)
    {
        p->update(dt);
        p->draw(m_render, m_frameData);
    }
    m_render->flush();
}

m3iTexture *m3cScene::loadTexture(const std::string &filename)
{
    if (filename.length() == 0)
        return 0;
    auto it = m_cachedTextures.find(filename);
    if (it != m_cachedTextures.end())
        return it->second;
    auto texture = m3cTextureLoader::loadTexture(m_render, m_system->getFileSystem(), filename);
    m_cachedTextures[filename] = texture;
    return texture;
}

void m3cScene::setShadowMapScale(float scale)
{
    m_frameData.shadowsData.scale = scale;
}

void m3cScene::setShadowsEnabled(bool enabled)
{
    m_frameData.shadowsData.enabled = enabled;
}

void m3cScene::setLightingEnabled(bool enabled)
{
    m_frameData.lighting = enabled;
}

void m3cScene::registerMovable(m3iMovable *movable)
{
    m_registeredMovables.push_back(movable);
}

void m3cScene::unregisterMovable(m3iMovable *movable)
{
    if (m_attachedPhysicsBodies.find(movable) != m_attachedPhysicsBodies.end())
        removedPhysicsBody(m_attachedPhysicsBodies[movable]);

    if (m_dontUnregisterMovables)
        return;

    remove_from_vector(m_registeredMovables, movable);
    // remove model is exists
    for (auto d = m_modelsForDraw.begin(); d != m_modelsForDraw.end(); ++d)
    {
        if (reinterpret_cast<const void*>(*d) ==
            reinterpret_cast<const void*>(movable))
        {
            m_modelsForDraw.erase(d);
            return;
        }
    }
    // remove particles is exists
    for (auto d = m_particlesForDraw.begin(); d != m_particlesForDraw.end(); ++d)
    {
        if (reinterpret_cast<const void*>(*d) ==
            reinterpret_cast<const void*>(movable))
        {
            m_particlesForDraw.erase(d);
            return;
        }
    }
    // remove batch is exists
    for (auto b = m_batchesForDraw.begin(); b != m_batchesForDraw.end(); ++b)
    {
        if (reinterpret_cast<const void*>(*b) ==
            reinterpret_cast<const void*>(movable))
        {
            m_batchesForDraw.erase(b);
            return;
        }
    }
}

uint64_t m3cScene::attachPostEffect(const std::string &fragmentSource, uint16_t width, uint16_t height)
{
    m_render->usePostEffectSet(m_postEffectSet);
    return m_render->attachPostEffect(fragmentSource, width, height);
}

void m3cScene::detachPostEffect(uint64_t id)
{
    m_render->usePostEffectSet(m_postEffectSet);
    m_render->detachPostEffect(id);
}

m3iPhysicsBody *m3cScene::attachPhysicsBoxBody(m3iMovable *movable, float mass, const glm::vec3 &dims)
{
    auto b = m_physicsWorld->createBoxBody(mass, dims);
    m_attachedPhysicsBodies[movable] = static_cast<m3cPhysicsBody*>(b);
    static_cast<m3cPhysicsBody*>(b)->setModelMatrix(movable->getModelMatrix(true));
    return b;
}

m3iPhysicsBody *m3cScene::attachPhysicsSphereBody(m3iMovable *movable, float mass, float radius)
{
    auto b = m_physicsWorld->createSphereBody(mass, radius);
    m_attachedPhysicsBodies[movable] = static_cast<m3cPhysicsBody*>(b);
    static_cast<m3cPhysicsBody*>(b)->setModelMatrix(movable->getModelMatrix(true));
    return b;
}

m3iPhysicsBody *m3cScene::attachPhysicsMeshBody(m3iModel *model, float mass)
{
    auto b = m_physicsWorld->createMeshBody(mass, static_cast<m3cModel*>(model)->getMesh());
    m_attachedPhysicsBodies[model] = static_cast<m3cPhysicsBody*>(b);
    static_cast<m3cPhysicsBody*>(b)->setModelMatrix(model->getModelMatrix(true));
    return b;
}

void m3cScene::setSunColor(const glm::vec4 &color)
{
    m_frameData.sun.color = color;
}

void m3cScene::setSunPosition(const glm::vec3 &position)
{
    m_frameData.sun.position = position;
}

void m3cScene::setSunDirection(const glm::vec3 &direction)
{
    m_frameData.sun.direction = direction;
}

void m3cScene::setFogParams(float near, float far, const glm::vec4 &color)
{
    m_frameData.fog.start = near;
    m_frameData.fog.end = far;
    m_frameData.fog.color = color;
}

void m3cScene::setFogEnabled(bool enabled)
{
    m_frameData.fog.enabled = enabled;
}

bool m3cScene::isFogEnabled() const
{
    return m_frameData.fog.enabled;
}

void m3cScene::setClearColor(const glm::vec4 &color)
{
    m_frameData.clearColor = color;
}

void m3cScene::addedPhysicsBody(m3iPhysicsBody *body)
{
}

void m3cScene::removedPhysicsBody(m3iPhysicsBody *body)
{
    for (auto it = m_attachedPhysicsBodies.begin(); it != m_attachedPhysicsBodies.end(); ++it)
    {
        if (it->second == body)
        {
            m_attachedPhysicsBodies.erase(it);
            return;
        }
    }
}

}
