#include "m3cSkeleton.h"

namespace melissa
{

void m3sSkeletonJointFrame::setPosition(float x, float y, float z){
    pos[0] = x; pos[1] = y; pos[2] = z;
}

void m3sSkeletonJointFrame::setRotation(float x, float y, float z, float w){
    rot[0] = x; rot[1] = y; rot[2] = z; rot[3] = w;
}

m3sSkeletonJoint::m3sSkeletonJoint()
    : parent(0)
{}

void m3sSkeletonJoint::sortFrames(){
    qsort(frames.data(), frames.size(), sizeof(m3sSkeletonJointFrame), [](const void *jf1, const void *jf2){
        return static_cast<const m3sSkeletonJointFrame*>(jf1)->time - static_cast<const m3sSkeletonJointFrame*>(jf2)->time;
    });
}

void m3sSkeletonJoint::setPosition(float x, float y, float z){
    bindPos[0] = x; bindPos[1] = y; bindPos[2] = z;
}

void m3sSkeletonJoint::setRotation(float x, float y, float z, float w){
    bindRot[0] = x; bindRot[1] = y; bindRot[2] = z; bindRot[3] = w;
}

void m3sSkeletonJoint::setParent(m3sSkeletonJoint *parent){
    if (this->parent)
        this->parent->children.erase(std::remove(this->parent->children.begin(), this->parent->children.end(), this), this->parent->children.end());

    this->parent = parent;

    if (this->parent)
        this->parent->children.push_back(this);
}

m3sSkeletonJoint *m3sSkeletonJoint::getParent() const{
    return this->parent;
}

std::vector<m3sSkeletonJoint *> &m3sSkeletonJoint::getChildren(){
    return children;
}

const std::vector<glm::mat4> &m3cSkeleton::getJoints(double frameTime)
{
    setFrame(frameTime);
    return m_finalMats;
}

uint64_t m3cSkeleton::getJointsCount() const
{
    return getCountJoints();
}

m3cSkeleton::m3cSkeleton()
    : m_needRebuild(true)
{}

void m3cSkeleton::build(const m3sSkeletonData &data)
{
    for (const m3sSkeletonData::Joint &j : data.joints){
        m3sSkeletonJoint joint;
        joint.bindPos = glm::vec3(j.pos[0], j.pos[1], j.pos[2]);
        joint.bindRot.x = j.quat[0];
        joint.bindRot.y = j.quat[1];
        joint.bindRot.z = j.quat[2];
        joint.bindRot.w = j.quat[3];
        for (const m3sSkeletonData::Frame &f : j.frames){
            m3sSkeletonJointFrame frame;
            frame.pos = glm::vec3(f.pos[0], f.pos[1], f.pos[2]);
            frame.rot.x = f.quat[0];
            frame.rot.y = f.quat[1];
            frame.rot.z = f.quat[2];
            frame.rot.w = f.quat[3];
            frame.time = f.time;
            joint.frames.push_back(frame);
        }
        this->m_joints.push_back(joint);
    }
    for (int i = 0; i < data.joints.size(); ++i){
        if (data.joints[i].parentIdx == -1)
            continue;

        m_joints[i].setParent(&m_joints[data.joints[i].parentIdx]);
    }
}

void m3cSkeleton::addJoint(const m3sSkeletonJoint &joint){
    m_joints.push_back(joint);
}

m3sSkeletonJoint &m3cSkeleton::getJoint(uint32_t index){
    return m_joints[index];
}

uint32_t m3cSkeleton::getCountJoints() const{
    return m_joints.size();
}

void m3cSkeleton::clear(){
    m_joints.clear();
}

bool m3cSkeleton::needRebuild() const{
    return m_needRebuild;
}

void m3cSkeleton::setAsRebuilt(){
    m_needRebuild = false;
}

void m3cSkeleton::setAsNeedRebuild(){
    m_needRebuild = true;
}

void m3cSkeleton::setFrame(double frameTime)
{
    m_finalMats.clear();

    if (needRebuild()){
        // build local pose matrices
        for (uint32_t i = 0; i < getCountJoints(); ++i){
            m3sSkeletonJoint &joint = getJoint(i);
            glm::mat4 poseMat = glm::toMat4(joint.bindRot);
            poseMat[3].x = joint.bindPos[0]; poseMat[3].y = joint.bindPos[1]; poseMat[3].z = joint.bindPos[2];
            joint.localMat = poseMat;
        }

        // calculate global matrices
        for (uint32_t i = 0; i < getCountJoints(); ++i){
            m3sSkeletonJoint &joint = getJoint(i);
            glm::mat4 globInvMat = glm::inverse(calculateGlobalMatrix(joint));
            joint.globInvMat = globInvMat;
        }

        setAsRebuilt();
    }

    // calculate frame matrices
    for (uint32_t i = 0; i < getCountJoints(); ++i){
        m3sSkeletonJoint &joint = getJoint(i);
        if (!joint.getParent())
            calculateFramesMatrix(true, joint, glm::mat4(1.0), frameTime);
    }

    // calculate final matrix
    for (uint32_t i = 0; i < getCountJoints(); ++i){
        m3sSkeletonJoint &joint = getJoint(i);
        m_finalMats.push_back(joint.globFrameMat * joint.globInvMat);
    }
}

void m3cSkeleton::calculateFramesMatrix(bool root, m3sSkeletonJoint &joint, const glm::mat4 &parentMatrix, double frameTime)
{
    // TODO: 56% optimize
    int currentFrame = -1;
    int nextFrame = -1;

    // find current frame
    for (int i = joint.frames.size() - 1; i >= 0; --i)
        if (frameTime > joint.frames[i].time){
            currentFrame = i;
            break;
        }

    nextFrame = currentFrame + 1;

    if (currentFrame == -1)
        currentFrame = 0;

    if (nextFrame >= joint.frames.size())
        nextFrame = currentFrame;

    m3sSkeletonJointFrame &frame1 = joint.frames[currentFrame];
    m3sSkeletonJointFrame &frame2 = joint.frames[nextFrame];

    // get interpolation scale
    float a = frame2.time - frame1.time;
    float v = frameTime - frame1.time;
    float s = a != 0 ? v * 1.0 / a : 0;
    if (s < 0) s = 0;
    if (s > 1) s = 1;

    // build interpolated matrix
    glm::vec3 locFramePos = glm::mix(frame1.pos, frame2.pos, s);
    glm::quat locFrameRot = glm::mix(frame1.rot, frame2.rot, s);
    glm::mat4 locFrameMat = glm::toMat4(locFrameRot);
    locFrameMat[3].x = locFramePos[0]; locFrameMat[3].y = locFramePos[1]; locFrameMat[3].z = locFramePos[2];

    joint.globFrameMat = (root ? (joint.localMat * locFrameMat) : (parentMatrix * (joint.localMat * locFrameMat)));

    for (int i = 0; i < joint.getChildren().size(); ++i){
        m3sSkeletonJoint *child = joint.getChildren().at(i);
        calculateFramesMatrix(false, *child, joint.globFrameMat, frameTime);
    }
}

glm::mat4 m3cSkeleton::calculateGlobalMatrix(m3sSkeletonJoint &joint)
{
    if (!joint.getParent())
        return joint.localMat;

    return calculateGlobalMatrix(*(joint.getParent())) * joint.localMat;
}

float *m3cSkeleton::getRawFinalMatrices()
{
    // prepared for shader data
    return reinterpret_cast<float*>(m_finalMats.data());
}

void m3cSkeleton::render(m3sSkeletonJoint *joint)
{
//    glDisable(GL_DEPTH_TEST);
//    glDisable(GL_BLEND);

//    glUseProgram(0);

//    glMatrixMode(GL_PROJECTION);
//    glLoadMatrixf(glm::value_ptr(glm::perspective(45.0f, 1.6f, 1.0f, 300.0f)));

//    glMatrixMode(GL_MODELVIEW);
//    glLoadMatrixf(glm::value_ptr(glm::lookAt(glm::vec3(5, 2, 2), glm::vec3(0, 0, 0), glm::vec3(0, 1, 0))));

//    if (joint->getParent()){
//        glPointSize(4);
//        glBegin(GL_POINTS);
//        glColor3f(1.0, 0.0, 0.0);
//        glVertex3f(joint->getParent()->globFrameMat[3].x, joint->getParent()->globFrameMat[3].y, joint->getParent()->globFrameMat[3].z);
//        glEnd();

//        glBegin(GL_LINES);
//        glColor3f(1.0, 0.0, 0.0);
//        glVertex3f(joint->getParent()->globFrameMat[3].x, joint->getParent()->globFrameMat[3].y, joint->getParent()->globFrameMat[3].z);
//        glVertex3f(joint->globFrameMat[3].x, joint->globFrameMat[3].y, joint->globFrameMat[3].z);
//        glEnd();
//    }
//    else
//    {
//        glPointSize(6);
//        glBegin(GL_POINTS);
//        glColor3f(1.0, 0.0, 0.0);
//        glVertex3f(joint->globFrameMat[3].x, joint->globFrameMat[3].y, joint->globFrameMat[3].z);
//        glEnd();
//    }

//    for (int i = 0; i < joint->getChildren().size(); ++i){
//        render(joint->getChildren()[i]);
//    }
}

void m3cSkeleton::render()
{
    for (int i = 0; i < m_joints.size(); ++i){
        if (!m_joints[i].getParent())
            render(&m_joints[i]);
    }
}

const m3sSkeletonJoint &m3cSkeleton::getJoint(uint32_t index) const{
    return m_joints[index];
}

}
