#pragma once

#include "m3hStd.h"
#include "m3hGlm.h"
#include "m3sModelSet.h"

namespace melissa
{

class m3iBatch;
class m3iLight;
class m3iModel;
class m3iCamera;
class m3iMovable;
class m3iTexture;
class m3iParticles;
class m3iPhysicsBody;
class m3iBatchDrawDelegate;
class m3iParticlesBehaviour;

class m3iScene
{
public:
    virtual ~m3iScene() {}

    /**
     * @brief createBatch movable element in 3d scene, and not movable in 2d scene
     * @param spaceSize space size for 2d render for example as screen size
     * @param filename texture for batching
     * @param drawDel delegate for draw batch elements
     * @param max elements in batch
     * @param is2d enable 2d mode ( disable depth, disable movable, enable space size )
     */
    virtual m3iBatch *createBatch(const glm::vec2 &spaceSize, const std::string &filename, m3iBatchDrawDelegate *drawDel, uint64_t max, bool is2d) = 0;

    //! create movable dummy
    virtual m3iMovable *createMovable() = 0;

    //! load models from json file, if already loaded, gets from cache
    virtual std::vector<m3sModelSet> loadModelSets(const std::string &filename) = 0;

    //! set active camera
    virtual void setCamera(m3iCamera *camera) = 0;

    //! create new camera and may be set as current
    virtual m3iCamera *createCamera(bool setAsCurrent) = 0;

    //! get current active camera
    virtual m3iCamera *getCamera() const = 0;

    //! create new light
    virtual m3iLight *createLight() = 0;

    /**
     * @brief createParticles create particle system
     * @param behav callbacks class for manage particle units
     * @param filename texture
     * @param count units
     */
    virtual m3iParticles *createParticles(m3iParticlesBehaviour *behav, const std::string &filename, uint64_t count) = 0;

    /**
     * @param fragmentSource
     * @param width texture, if zero then size as screen size
     * @param height texture, if zero then size as screen size
     * @return post effect's id
     */
    virtual uint64_t attachPostEffect(const std::string &fragmentSource, uint16_t width, uint16_t height) = 0;

    //! detach post effect by id
    virtual void detachPostEffect(uint64_t id) = 0;

    //! for big scene set shadowmap scale
    virtual void setShadowMapScale(float scale) = 0;

    virtual void setShadowsEnabled(bool enabled) = 0;
    virtual void setLightingEnabled(bool enabled) = 0;

    virtual m3iTexture *loadTexture(const std::string &filename) = 0;

    virtual m3iPhysicsBody *attachPhysicsBoxBody(m3iMovable *movable, float mass, const glm::vec3 &dims) = 0;
    virtual m3iPhysicsBody *attachPhysicsSphereBody(m3iMovable *movable, float mass, float radius) = 0;
    virtual m3iPhysicsBody *attachPhysicsMeshBody(m3iModel *model, float mass) = 0;

    virtual void setSunColor(const glm::vec4 &color) = 0;
    virtual void setSunPosition(const glm::vec3 &position) = 0;
    virtual void setSunDirection(const glm::vec3 &direction) = 0;

    //! fog near and far must be clamped [0 - 1]
    virtual void setFogParams(float near, float far, const glm::vec4 &color) = 0;
    virtual void setFogEnabled(bool enabled) = 0;
    virtual bool isFogEnabled() const = 0;

    virtual void setClearColor(const glm::vec4 &color) = 0;
};

}
